#ifndef	LF_UTIL_H
#define	LF_UTIL_H

/* files included */
#include <stdio.h>
#include <math.h>
#ifndef PI
#define PI M_PI
#endif
#ifndef min
#define min(x,y) ((x) < (y) ? (x) : (y))
#endif
#ifndef max
#define max(x,y) ((x) < (y) ? (y) : (x))
#endif
#ifdef STATS
#include <sys/time.h>
#endif /* STATS */
#include "common.h"
#include "geom.h"
#include "lf_consts.h"

/*
** vector tools from volpack - use copied version if volpack.a is not used.
*/
#ifndef NO_VP
#include <volpack.h>
#else
typedef double vpVector3[3];		/* 3 element vector */
typedef double vpVector4[4];		/* 4 element vector */
typedef double vpMatrix3[3][3];		/* 3 by 3 element matrix */
typedef double vpMatrix4[4][4];		/* 4 by 4 element matrix */
typedef void * vpContext;

extern void		    vpIdentity3 (vpMatrix3 m);
extern void		    vpIdentity4 (vpMatrix4 m);
extern void		    vpNormalize3 (vpVector3 v);
extern void		    vpMatrixVectorMult4 (vpVector4 v2,
				vpMatrix4 m, vpVector4 v1);
extern void		    vpMatrixMult4 (vpMatrix4 m3,
				vpMatrix4 m2, vpMatrix4 m1);
extern void		    vpCrossProduct (vpVector3 p,
				vpVector3 v, vpVector3 w);
/* set fields of a vector */
#define vpSetVector3(v, v0, v1, v2)	v[0]=v0; v[1]=v1; v[2]=v2
#define vpSetVector4(v, v0, v1, v2, v3)	v[0]=v0; v[1]=v1; v[2]=v2; v[3]=v3

#endif
#define vpDiffVector3(target, first, second) \
        vpSetVector3(target, first[0] - second[0], \
		     first[1] - second[1], \
		     first[2] - second[2])
#define vpAddVector3(target, first, second) \
        vpAddVector3(target, first[0] + second[0], \
		     first[1] + second[1], \
		     first[2] + second[2])
#define vpDotProduct3(v, p) (v[0]*p[0] + v[1]*p[1] + v[2]*p[2])

/* definitions */

/* tag list definitions;
   IMO, this is broken by design:
   the number of arguments for a tag is varying,
   so replicaList() interprets every entry in the tag
   list as tag even if it is an argument =>
   you cannot pass arguments that are equal to LF_NULL,
   and whereever a tag list is scanned, all possible
   tags have to be known to skip over their arguments.
   Pointers cannot be passed at all. */

#define LF_NULL   (-1000)
#define LF_IGNORE (-2000)


/* definitions related to different lightfield operations */
#define	LF_OP_DRAW	0
#define	LF_OP_CREATE	1
#define	LF_OP_COMPRESS	2
#define	LF_OP_WRITE	3
#define	LF_OP_READ	4
#define	LF_OP_MIN	LF_OP_DRAW
#define	LF_OP_MAX	LF_OP_READ
#define	LF_OP_COMPRESS_TRAIN	5
#define	LF_OP_MEM_INPUT	6
#define	LF_OP_SHUFFLE	7

#define	LF_DRAW		LF_OP_DRAW

/* definitions related to drawing description */
#define	LF_DRAW_FILTER	0x1
#define	LF_DRAW_SCANLINE_CLAMP_RATIO 0x2
#define	LF_DRAW_TIMING_USER	0x3
#define	LF_DRAW_TIMING_SYSTEM	0x4
#define	LF_DRAW_STATS_RESET 0x5

/* definitions related to slab culling and intersection */
#define	LF_CULL_NO	0
#define	LF_CULL_1	1
#define	LF_CULL_2	2
#define	LF_CULL_3	3
#define	LF_CULL_4	4
#define	LF_NO_INTERSECT	5

/* definitions related to data flow of an internal operator chain */
#define	LF_SRC_SLABMEM	0
#define	LF_SRC_FILE	1
#define	LF_SRC_CREATE	2
#define	LF_SRC_DFMEM	3

#define	LF_DST_SLABMEM	0
#define	LF_DST_FILE	1
#define	LF_DST_SLABVQ	2
#define	LF_DST_FRAMEBUF	3
#define	LF_DST_DFMEM	4

#define LF_DF_CREATE_COMPRESS		0x0001
#define LF_DF_CREATE_WRITE		0x0002
#define LF_DF_COMPRESS_WRITE		0x0004
#define LF_DF_READ_COMPRESS		0x0008
#define LF_DF_CREATE_COMPRESS_WRITE	0x0010
#define LF_DF_READ_COMPRESS_WRITE	0x0020
#define LF_DF_DRAW			0x0040
#define LF_DF_CREATE			0x0080
#define LF_DF_COMPRESS			0x0100
#define LF_DF_WRITE			0x0200
#define LF_DF_READ			0x0400
#define	LF_DF_CREATE_COMPRESS_TRAIN	0x0800
#define	LF_DF_READ_COMPRESS_TRAIN	0x1000
#define	LF_DF_COMPRESS_TRAIN		0x2000

#define	LF_DF_FIELD	0
#define	LF_DF_SLAB	1
#define	LF_DF_SLICE_ST	2

/* handy aliases for addressing dataflow position array */
#define	LF_SLAB_POS	0
#define	LF_U_POS	1
#define	LF_V_POS	2
#define	LF_S_POS	3
#define	LF_T_POS	4
#define	VV_POS		5
#define UU_POS          6
#define	UVP_POS		7
#define LF_P_POS(i)     (8 + (i))
#define LF_PP_POS(i)    (LF_P_POS(LF_MAXPARAMS) + i)
#define LF_MAX_POS      LF_PP_POS(LF_MAXPARAMS)
int lfSum(const int *vals, int start, int end);
int lfMul(const int *vals, int start, int end);
#define lfMulPPos(pos) lfMul(pos, LF_P_POS(0), LF_P_POS(LF_MAXPARAMS))

/* definition related to filtering */
#define	LF_LERP_UV	0x1
#define LF_LERP_ST	0x2

/* definitions related to parameterisation */
#define LF_MAXPARAMS     4
typedef struct {
  int min, max;          /* parameter is evenly spaced between these two values */
  char *descr;           /* some name for it */
  int current;           /* current value */
  int def;               /* and its default */
} LFParam;
typedef struct {
  int num;
  LFParam params[LF_MAXPARAMS];
} LFParams;

/* definitions related to integrated renderer(s) */
enum LF_RendererType {
  LF_RENDERER_NONE,
  LF_RENDERER_VOLPACK,
  LF_RENDERER_VOLVIS,
  LF_RENDERER_RENDERMAN,
  LF_RENDERER_DUMMY
};
typedef struct {
  enum LF_RendererType type;
  int flags;
#define LFRF_NO_RENDERING (1<<0)
  int sup_qualities;      /* bit mask */
  int quality;            /* requested quality */
  char *config;
  char *lidDir;
  /* renderer can accept this number of arguments
     that are taken from the parameter array */
  int num_args;
  int arg_indices[LF_MAXPARAMS];

  void *handle;   /* private handle of renderer, allocated with lfRDAlloc() */
  float aspect;   /* prefered x/y value for lfRDRenderParams() */
} LFRenderer;
bool_t lfRDAlloc(LFRenderer *renderer);
void lfRDDone(LFRenderer *renderer);
bool_t lfRDRender(LFRenderer *renderer,
		   Matrix4 mat_model, Matrix4 mat_proj, int width, int height,
		   int format, void *pixels);
/*
** render parameters using OpenGL and
** coordinates from 0.0/0.0 to 1.0/1.0
*/
void lfRDRenderParams(LFRenderer *renderer);


/* padding definitions */
#define	LF_PAD_FIRST	0x0
#define	LF_PAD_LAST	0x1

/* types */
typedef struct {
    short u, v, s, t;
} LFPixelUVST;

#ifdef ABGR_LF
typedef struct {
    unsigned char a, b, g, r;
} LFPixelRGBA8;
typedef struct {
    unsigned char b, g, r;
} LFPixelRGB8;
#else
typedef struct {
    unsigned char r, g, b, a;
} LFPixelRGBA8;
typedef struct {
    unsigned char r, g, b;
} LFPixelRGB8;
#endif
typedef int LFInt32;


typedef struct {
    float ox, oy, oz, ow;
    float x, y, z, w;
    float s, t;
} LFVtx;

typedef struct {
    float x, y;		/* screen space vertex coordinates */
    /* here w and q are really 1/w and 1/q */
    float z1, w;	/* z and w with respect to quad UV */
    float z2, q;	/* z and w with respect to quad ST */
    /* here u, v, s, and t are really u/w, v/w, s/q, and t/q */
    float u, v, s, t;
} LFScreenVtx;

typedef struct {
    int id;
    int size;			/* size of the code book */
    int ubits, vbits, sbits, tbits, pbits[LF_MAXPARAMS];
    int intern_format;		/* internal format for code words */
    float train_pct;		/* percentage of training set */
    void *codebook;
    int type;			/* what type of data in codebook */
#define	LF_VQ_LINEAR_DATA	0x0
#define	LF_VQ_LINEAR_VQ		0x1
#define	LF_VQ_TREE_VQ		0x2
    int sample_size;		/* per [u,v,s,t, pi] sample size */
}  LFVQCodebook;
#define lfSumPBits(vq) lfSum(vq->pbits, 0, LF_MAXPARAMS)

/*
 * Arbitrary sampling of the additional parameters is possible.
 * The number and values of a certain parameter are reached by
 * following the indices of the previous parameters.
 */
struct LFParamSamples;
typedef struct {
  int p;                    /* value of the current parameter */
  struct LFParamSamples *psamples; /* number and values for following parameters
			       or NULL if last parameter */
} LFSample;
typedef struct LFParamSamples {
  int np;                  /* number of samples for current parameter */
  LFSample samples[1];     /* (np - 1) entries follow */
} LFParamSamples;
int lfVQIndexTileStart(int num_params, int pbits[LF_MAXPARAMS], int j[LF_MAXPARAMS], LFParamSamples *psamples, int tile_start[LF_MAXPARAMS]);
#define lfVQIndex(num_params, pbits, j, psamples) lfVQIndexTileStart(num_params, pbits, j, psamples, NULL)
int lfMaxVQIndex(int num_params, int pbits[LF_MAXPARAMS], LFParamSamples *psamples);
int lfIndex(int num_params, int j[LF_MAXPARAMS], LFParamSamples *psamples);
int lfMaxIndex(int num_params, LFParamSamples *psamples);
void lfExtractParams(int num_params, const int p[LF_MAXPARAMS], LFParamSamples *psamples, int values[LF_MAXPARAMS]);
void lfWritePSamples(FILE *outFile, int first_indent, int level_indent,
		     int pnum, int num_params, LFParamSamples *psamples);
void lfCalcMaxNP(int num_params, LFParamSamples *psamples, int max_np[LF_MAXPARAMS]);

/*
 * This defines what sources are used in which way to obtain an image.
 */
#define LF_MAX_SOURCES (1<<LF_MAXPARAMS)
#define LF_MAX_INTERP  (1<<LF_MAXPARAMS)
typedef struct {
  int num_accesses;                               /* number of samples required for interpolation;
						     <= 2^LF_MAXPARAMS */
  int p_indices[LF_MAX_SOURCES][LF_MAXPARAMS];    /* generic light fields to be used for each of the
						     num_acceses samples required for interpolation */
  int num_interp;                                 /* number of interpolations required;
						     <= 2^LF_MAXPARAMS - 1 */
  /* the first interpolation to do is the one at the end */
  int source[LF_MAX_INTERP],                    /* samples are taken from source and target sample and stored */
    target[LF_MAX_INTERP];                      /* in target sample before proceeding to next interpolation */
  int fp[LF_MAX_INTERP];                        /* factor and bit shift for each parameter interpolation */
  int fp_bits[LF_MAX_INTERP];                   /* as required by LERP() macro */
#define FP_BITS 6                                 /* bits could also be fixed instead! */
} LFInterpolationPlan;


/*
 * the following struture contains information that might be 
 * shared by multiple slabs (for example, all slabs in a lightfield).
 */
typedef struct {
    /* current dimensions of u, v, s, t */
    int nu, nv, ns, nt;
    LFParamSamples *psamples;
    int max_np[LF_MAXPARAMS];    /* max_np[i] might not be reached everywhere if
				    samples were used! */
    /* remember if psamples has been created from a rectangular grid
       and from which one, as VQ encoding might need that */
    bool_t rectangular;          /* rectangular grid is stored in psamples */
    bool_t samples;              /* samples were also used */

    /*
     * In addition, the complete interpolation in an area of two parameters can be specified explicitely:
     */
    LFInterpolationPlan *plans;            /* an explicit specification of how to interpolate,
                                           indexed by [p1 - p1 min][p0 - p0 min ] */
#define PLAN(sh,params,p1,p0) sh->plans[(p1 - (params)->params[1].min) * \
				        ((params)->params[1].max - (params)->params[1].min + 1) + \
				        p0 - (params)->params[0].min]

    int u_frac_bits, v_frac_bits, s_frac_bits, t_frac_bits;
    int *uindex, *vindex, *sindex, *tindex;

    /* how (u, v, s, t) is translated into tile indices */
    int *vqtile_uindex, *vqtile_vindex, *vqtile_sindex, *vqtile_tindex;
    LFVQCodebook *vq;		/* vq codebook */
    int intern_format;		/* internal format for lightfield data */
    int sample_size;		/* per [u,v,s,t] sample size */
    int flag;			/* flag */
#define	LF_SLAB_VALID_AUX	0x1

    /* how to do the resampling */
    LFInterpolationPlan plan;
}  LFShared;
#define MAX_PINDEX(sh) lfMaxIndex(__LF_GET_PARAMS()->num, sh->psamples)
#define MAX_PVQINDEX(sh) lfMaxVQIndex(__LF_GET_PARAMS()->num, sh->vq->pbits, sh->psamples)


typedef struct {
    int id;
    LFVtx uv[4];
    LFVtx st[4];
    void *lightfield;
    int sample_size;
    LFShared *shared;

  /* for slab dependent parameter interpolation */
  void *lfs[LF_MAX_SOURCES];                     /* the subcubes of the lightfield and tile offsets */
  LFPixelRGBA8 *vqtiles_p[LF_MAX_SOURCES];       /* in the codebook corresponding to the parameters
				        	    of the shared->num_accesses samples */
} LFSlab;

typedef struct LFField {
    LFSlab *slabs[LF_MAXSLABS];
    int nslabs;
    unsigned enable_msk;	/* how many slabs are currently enabled */
    LFShared *shared;
    LFRenderer renderer;
    LFParams   params;
    int id;
    struct LFField *next;
} LFField;

#define	__LF_SLAB_ACTIVE(field,i)		\
	(field->enable_msk & (1 << i))
#define	__LF_SET_SLAB_ACTIVE(field,i,active)	\
	(field->enable_msk = (field->enable_msk	\
		& ~(1 << (i))) | ((active) << i))

typedef struct {
  /* external structure */
  int filter;
  int format;  /* LF_RGBA, LF_RGB */
  int type;
  bool_t interpolate[LF_MAXPARAMS];  /* parameters that may be interpolated */

  /*
  ** alternative method to specify viewing:
  */
  bool_t force_rendering;          /* do original rendering even if LF exists */
  bool_t rendering_done;           /* image is already finished */
  float fov;
  bool_t perspective;              /* do perspective projection */
  /* if fov == 0.0, then use explizit definition of screen instead: */
  float screen_d,                  /* distance between viewer and screen */
    screen_left, screen_right,     /* and its coordinates */
    screen_lower, screen_upper;

  int nu;                          /* number of sampling points in screen width */
  float n, f;                      /* distance between viewer and near and far
				      clip plane */
  float xobs, yobs, zobs;          /* coordinates of the observer */
  float xcenter, ycenter, zcenter; /* coordinates of what he's looking at */ 
  float yrot, xrot, zrot;          /* rotations after selecting this view */
  bool_t useIt;                    /* use these parameters to select view */

    /* internal structure */
    int pixel_size;
    int x, y;
    int width, height; 
    Matrix4 mat_model;
    Matrix4 mat_proj;
    Viewport viewport;
    LFPixelUVST *sample_buffer;
    void *pixel_buffer;
    int sample_buffer_size;
    int pixel_buffer_size;

  /* when nothing changed, then just use the old frame buffer */
  Matrix4 old_mat_model, old_mat_proj;
  Viewport old_viewport;
  int old_p[LF_MAXPARAMS];
  bool_t old_interpolate[LF_MAXPARAMS];
  bool_t old_force_rendering;
  int old_filter, old_quality;

#if 0
  /* when changing only one parameter, two images are kept and
     interpolation is done between them */
  LFPixelOutputRGB8 *images[2];
  int old_width, old_height, old_filter;
  int old_p[LF_MAXPARAM];
  int num_interp;
  bool_t fast_interp;
  int changed_p, old_changed_p;
#endif

#ifdef	STATS
    float draw_time_user, draw_time_system;
    int scanline_clamped;
    int scanline_total;
#endif
} LFView;

/* an operation applied to a lightfield */
typedef struct LFOp {
    int op_id;
    int op_method;
    float *op_descr;
    void *op_extra;
} LFOp;

#define	LF_MAX_OPS		10
#define	LF_MAX_INTERNOPS	LF_MAX_OPS

struct LFOps;			/* forward declaration */

typedef struct LFInternOp {
    int op_id;
    int op_method;
    float *op_descr;
    void *op_extra;

    /* really an operator on the data flow */
    void (*op_func)(LFSlab ** /*slabs*/, int /*nslab*/, const int * /*pos*/,
	const float * /*descr*/, const void * /*input*/, int /*input_size*/,
	void * /*output*/, int /*output_size*/, void * /*aux_ptr*/);

    /* functions to call after data flow ends */
    void (*op_end)(struct LFOps * /*ops*/, struct LFInternOp * /*op*/, void * /*aux_ptr*/);

    /*
     * update function for next loop, this hides the internal looping and
     * I/O mechanism of dataflow from the operator
     */
    void (*op_update)(LFSlab ** /*slabs*/, int /*nslab*/, struct LFInternOp * /*op*/);
    int op_loop;
    int op_pos[LF_MAX_POS + 1];

    /* pre-determined I/O memory and size for data flow of each operator */
    char *op_input_alloced;
    char *op_input;
    int op_input_size;
    char *op_output;
    int op_output_size;

    /* auxiliary data structure hanging off an internal operator */
    void *op_aux;
} LFInternOp;

/* stuff related to lfBegin/lfEnd */
typedef struct LFOps {
    /* stuff describing slabs involved for lfBegin/lfEnd */
    LFSlab *op_slabs[LF_MAXSLABS];
    int op_slab_cnt;

    /* stuff describing an external operator chain */
    float *op_hints;
    LFOp ops[LF_MAX_OPS];
    int op_cnt;

    /* stuff describing an internal operator chain */
    int chain_src;		/* LF_SRC_SLABMEM/LF_SRC_FILE/LF_SRC_CREATE/LF_SRC_MEM */
    int chain_dst;		/* LF_DST_SLABMEM/LF_DST_FILE/LF_SRC_MEM/LF_FRAMEBUF */
    int chain_flow;		/* LF_DF_VOL4D, LF_DF_SLICE_ST */
    int chain_type;		/* what type of chain is it */
    LFInternOp chain_ops[LF_MAX_INTERNOPS];
    int chain_cnt;
    int chain_pos[LF_MAX_POS];		/* current position in terms of source data */
    void *chain_input;		/* source memory */
    void *chain_output;		/* destination memory */
    void *chain_mem;		/* temporary memory in data-flow analysis */
    int chain_msk;		/* various mask */
#define	LF_SHARED		0x01
#define	LF_GEN_VQ_TRAINSET	0x02
#define	LF_GEN_VQ_CODEARRAY	0x04
#define	LF_SET_SLAB		0x08
#define	LF_NEW_SLAB		0x10
#define	LF_GEN_VQ_CODEBOOK	0x20
} LFOps;

typedef struct {
    LFOps ops;
    LFField *field;
    LFField *lf_list;
    LFView view;
#ifdef STATS
    struct itimerval timer_user;
    struct itimerval timer_system;
#endif
} LFContext;

typedef struct {
    void (*get_transform)(Matrix4 /*mat_model*/, Matrix4 /*mat_proj*/,
	int * /*x*/, int * /*y*/, int * /*width*/, int * /*height*/);
    void (*write_pixrect)(int /*x*/, int /*y*/, int /*w*/, int /*h*/,
	int /*format*/, int /*type*/, void * /*pixels*/);
} LFDriver;

/* macros */

#define	__LF_GET_CONTEXT()	__lf_context
#define	__LF_GET_DRIVER()	__lf_driver
#define __LF_GET_PARAMS()       (&__LF_GET_CONTEXT()->field->params)
#define __LF_GET_RENDERER()     (&__LF_GET_CONTEXT()->field->renderer)
#define __LF_GET_VIEW()         (&__LF_GET_CONTEXT()->view)

#ifdef ABGR_OUTPUT
typedef struct {
    unsigned char a, b, g, r;
} LFPixelOutputRGBA8;
typedef struct {
    unsigned char b, g, r;
} LFPixelOutputRGB8;
#else
typedef struct {
    unsigned char r, g, b, a;
} LFPixelOutputRGBA8;
typedef struct {
    unsigned char r, g, b;
} LFPixelOutputRGB8;
#endif


/* external variables */
extern LFContext	*__lf_context;
extern LFDriver		*__lf_driver;
extern LFSlab		*__lf_descr_slabs;
extern bool_t          lfVerbose;
extern bool_t          lfNoITiming;
extern float           *lfIntensity;   /* if set, drawing will calculate the final
					  pictures average intensity */
/* expects RGBA byte order: */
float lfCalcIntensity(unsigned char *pixels, int width, int height, int sample_size);
float lfCalcIntensityStride(unsigned char *pixels, int width, int height, int stride);

/* prototypes */
#if 0
/* obsolete */
extern LFField
    *lfReadField(char *name);
#endif
char *dirname(char *name);

extern int
    log2(int /*size*/);
extern bool_t 
    lfSlabIntersecting(LFVtx /*uv*/[4], LFVtx /*st*/[4],
	const Matrix4 /*model*/, const Matrix4 /*proj*/,
	LFScreenVtx * /*inter*/, int /*ninter*/[2]);

extern void
    *lfBeginCreate(LFOps * /*ops*/, LFInternOp * /*op*/),
    *lfBeginDrawFB(LFOps * /*ops*/, LFInternOp * /*op*/),
    *lfBeginMemShuffle(LFOps * /*ops*/, LFInternOp * /*op*/),
    *lfBeginReadLif(LFOps * /*ops*/, LFInternOp * /*op*/),
    *lfBeginReadSliceFile(LFOps * /*ops*/, LFInternOp * /*op*/),
    *lfBeginWriteLif(LFOps * /*ops*/, LFInternOp * /*op*/),
    lfDataflowAnalysis(LFOps * /*ops*/),
    lfDrawFB(LFSlab ** /*slabs*/, int /*nslab*/, const int * /*pos*/,
	const float * /*descr*/, const void * /*input*/, int /*input_size*/,
	void * /*output*/, int /*output_size*/, void * /*aux_ptr*/),
    lfDrawFieldFramebuffer(LFField * /*field*/),
    lfEndDefault(LFOps * /*ops*/, LFInternOp * /*op*/, void * /*aux_ptr*/),
    lfEndDrawFB(LFOps * /*ops*/, LFInternOp * /*op*/, void * /*aux_ptr*/),
    lfEndReadLif(LFOps * /*ops*/, LFInternOp * /*op*/, void * /*aux_ptr*/),
    lfEndReadSliceFile(LFOps * /*ops*/, LFInternOp * /*op*/, void * /*aux_ptr*/),
    lfEndVQCompress(LFOps * /*ops*/, LFInternOp * /*op*/, void * /*aux_ptr*/),
    lfEndWriteLif(LFOps * /*ops*/, LFInternOp * /*op*/, void * /*aux_ptr*/),
    lfError(char * /*fmt*/, ...),
    lfGetParameters(int /*op_name*/, int /*pname*/, float * /*param*/),
    lfLinearizeVQCodebook(LFVQCodebook * /*vq*/),
    lfMakeIndices(LFShared * /*shared*/),
    lfMemInputSlice(LFSlab ** /*slabs*/, int /*nslab*/, const int * /*pos*/,
	const float * /*descr*/, const void * /*input*/, int /*input_size*/,
	void * /*output*/, int /*output_size*/, void * /*aux_ptr*/),
    lfMemShuffle(LFSlab ** /*slabs*/, int /*nslab*/, const int * /*pos*/,
	const float * /*descr*/, const void * /*input*/, int /*input_size*/,
	void * /*output*/, int /*output_size*/, void * /*aux_ptr*/),
    lfOutput(char * /*fmt*/, ...),
    lfPadData(LFSlab * /*slab*/, int /*pad_type*/, int /*new_size*/, int /*mode*/),
    lfRasterizeTriPntSample(const LFSlab * /*slab*/, const LFScreenVtx * /*v0*/,
	const LFScreenVtx * /*v1*/, const LFScreenVtx * /*v2*/),
    lfReadLif(LFSlab ** /*slabs*/, int /*nslab*/, const int * /*pos*/,
	const float * /*descr*/, const void * /*input*/, int /*input_size*/,
	void * /*output*/, int /*output_size*/, void * /*aux_ptr*/),
    lfReadSliceFile(LFSlab ** /*slabs*/, int /*nslab*/, const int * /*pos*/,
	const float * /*descr*/, const void * /*input*/, int /*input_size*/,
	void * /*output*/, int /*output_size*/, void * /*aux_ptr*/),
    lfResampleRGBA8(const LFSlab * /*slab*/, LFPixelUVST * /*scanline*/,
	int /*x*/, int /*y*/, int /*w*/),
    lfSerialUpdateBlockPoint(LFSlab ** /*slabs*/, int /*nslab*/,
	LFInternOp * /*op*/),
    lfSerialUpdateBlockSlice(LFSlab ** /*slabs*/, int /*nslab*/,
	LFInternOp * /*op*/),
    lfSerialUpdatePoint(LFSlab ** /*slabs*/, int /*nslab*/, LFInternOp * /*op*/),
    lfSetParameters(int /*op_name*/, float * /*param_list*/),
    lfStartTimer(void),
    lfEndTimer(void),
    lfUpdateNoop(LFSlab ** /*slabs*/, int /*nslab*/, LFInternOp * /*op*/),
    lfVQGenCodebook(LFSlab ** /*slabs*/, int /*nslab*/, const int * /*pos*/,
	const float * /*descr*/, const void * /*input*/, int /*input_size*/,
	void * /*output*/, int /*output_size*/, void * /*aux_ptr*/),
    lfVQCompress(LFSlab ** /*slabs*/, int /*nslab*/, const int * /*pos*/,
	const float * /*descr*/, const void * /*input*/, int /*input_size*/,
	void * /*output*/, int /*output_size*/, void * /*aux_ptr*/),
    lfWriteLif(LFSlab ** /*slabs*/, int /*nslab*/, const int * /*pos*/,
	const float * /*descr*/, const void * /*input*/, int /*input_size*/,
	void * /*output*/, int /*output_size*/, void * /*aux_ptr*/);


#include "lf_funcs.h"
#include "fileIO.h"

#endif	/* ! LF_UTIL_H */

