%{
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "y.tab.h"

int yylinenumber = 1;
void skip_comments(void);
void skip_to_the_end(void);
%}
alpha   [a-zA-Z]
special	[\.\_-]
digit	[0-9]
space   [ \t\n]
string	 ({alpha}|{digit}|{special})*
spacestr ({alpha}|{digit}|{special}|[ \t])*
filepath "/"?"/"?(("."|".."|{string})"/")*{string}
quotedfp \""/"?"/"?(("."|".."|{spacestr})"/")*{spacestr}\"

%%

#		skip_comments();
FIELD_DONE	skip_to_the_end();
RENDERER        return RENDERER;
VOLPACK         return VOLPACK;
VOLVIS          return VOLVIS;
RENDERMAN       return RENDERMAN;
DUMMY           return DUMMY;
QUALITY         return QUALITY;
PARAMETER       return PARAMETER;
SLAB		return SLAB;
FORMAT		return FORMAT;
SLAB_UV		return SLAB_UV;
SLAB_ST		return SLAB_ST;
SAMPLES_UV	return SAMPLES_UV;
SAMPLES_ST	return SAMPLES_ST;
SAMPLES_P       return SAMPLES_P;
SAMPLE_POINTS_P return SAMPLE_POINTS_P;
SAMPLE_POINT    return SAMPLE_POINT;
INTERPOLATION   return INTERPOLATION;
RGBA		return RGBA;
RGB		return RGB;
[+-]?{digit}*\.{digit}+/{space} { yylval.f = atof(yytext); return REAL; }
[+-]?{digit}+/{space}   	{ yylval.i = atoi(yytext); return INTEGER; }
{filepath}              	{ yylval.s = strdup(yytext); return FILEPATH; }
{quotedfp}                      { yylval.s = strdup(&yytext[1]); yylval.s[strlen(yylval.s) - 1] = 0; return FILEPATH; }
[ \t]		;
\n              yylinenumber++;

%%
int
yywrap()
{
    return 1;
}

void
skip_comments(void)
{
    while (input() != '\n');
    yylinenumber++;
}

void
skip_to_the_end(void)
{
    int c;

    while ((c=getchar()) != EOF) {
       if(c == '\n') {
          yylinenumber++;
       }
    }
    ungetc(c, stdin);
}
