/*
 * draw_fb.c
 *
 * this file eventually will replace draw.c
 */
#include "lightfield.h"
#include "draw.h"
#include "renderer.h"
#include <string.h>
#include <strings.h>


/* setup interpolation variables:
   recursively go through all parameters and find out
   how to sample and/or interpolate them */
static int
interpolateParam(LFShared *shared, int pnum, int interp_level,
		 LFParamSamples *psamples, int p_indices[LF_MAXPARAMS])
{
  LFParams *params = __LF_GET_PARAMS();
  LFView *view = &__LF_GET_CONTEXT()->view;
  int e;
  int local_target = shared->plan.num_accesses;

  /* don't do further sampling/interpolation, use the last result */
  if(pnum >= params->num) {
    memcpy(shared->plan.p_indices[shared->plan.num_accesses], p_indices, sizeof(int) * LF_MAXPARAMS);
    shared->plan.num_accesses++;
  } else {
    int cur_p = params->params[pnum].current;

    /* find closest match for nearest neighbor or left neighbor for interpolation */
    for(e = 0; e < psamples->np - 1; e++) {
      if(view->interpolate[pnum] ?
	 psamples->samples[e + 1].p > cur_p /* interpolate between e and e + 1 */
	 : abs(psamples->samples[e].p - cur_p) <= abs(psamples->samples[e + 1].p - cur_p) /* nearest neighbor */)
	break;
    }

    /* interpolate if allowed,
       not exactly sampled and
       in the middle of two sampling points */
    if(view->interpolate[pnum]
       && abs(psamples->samples[e].p - cur_p)
       && e < psamples->np - 1
       && !(e == 0 && cur_p < psamples->samples[0].p)) {
      /* sample lightfield for left and right value and interpolate */
      p_indices[pnum] = e;
      shared->plan.target[shared->plan.num_interp] =
	interpolateParam(shared, pnum + 1, interp_level + 1,
			 psamples->samples[e].psamples, p_indices);
      assert(shared->plan.target[shared->plan.num_interp] == local_target);
      p_indices[pnum] = e + 1;
      shared->plan.source[shared->plan.num_interp] =
	interpolateParam(shared, pnum + 1, interp_level + 1,
			 psamples->samples[e + 1].psamples, p_indices);
      assert(shared->plan.source[shared->plan.num_interp] > local_target);
      shared->plan.fp_bits[shared->plan.num_interp] = FP_BITS;
      shared->plan.fp[shared->plan.num_interp] = (cur_p - psamples->samples[e].p)
	* (1<<shared->plan.fp_bits[shared->plan.num_interp])
	/ (psamples->samples[e + 1].p - psamples->samples[e].p);
      shared->plan.num_interp++;
    } else {
      int source;

      /* simply sample at the current position */
      p_indices[pnum] = e;
      source = interpolateParam(shared, pnum + 1, interp_level,
				psamples->samples[e].psamples, p_indices);
      assert(source == local_target);
    }
  }
  return local_target;
}

/* test if all parameters are allowed to be interpolated */
static int
interpolateAllParams(void)
{
  LFParams *params = __LF_GET_PARAMS();
  LFView *view = &__LF_GET_CONTEXT()->view;
  int i;

  for(i = 0; i < params->num; i++) {
    if(!view->interpolate[i]) {
      return FALSE;
    }
  }
  return TRUE;
}


static void
setInterpolationParams(LFSlab *slab)
{
  LFShared *sh = slab->shared;
  LFParams *params = __LF_GET_PARAMS();
  int pindices[LF_MAXPARAMS];
  int i, index;

  if(sh->plans && interpolateAllParams()) {
    int index = 0;

    for(i = params->num - 1; i >= 0;) {
      int cur = params->params[i].current;

      if(cur < params->params[i].min) {
	cur = params->params[i].min;
      }
      if(cur > params->params[i].max) {
	cur = params->params[i].max;
      }
      index += cur - params->params[i].min;
      if(--i >= 0) {
	index *= params->params[i].max - params->params[i].min + 1;
      }
    }
    sh->plan = sh->plans[index];
  } else {
    sh->plan.num_accesses = 0;
    sh->plan.num_interp   = 0;
    interpolateParam(sh, 0, 0, sh->psamples, pindices);
    /* reverse order of interpolation so that access is faster later
       because the loop can count down to 0 */
    for(i = 0; i < sh->plan.num_interp/2; i++) {
      int tmp;
#define SWAP(x) tmp = x[i], x[i] = x[sh->plan.num_interp - 1 - i], x[sh->plan.num_interp - 1 - i] = tmp
      SWAP(sh->plan.source);
      SWAP(sh->plan.target);
      SWAP(sh->plan.fp);
      SWAP(sh->plan.fp_bits);
    }
  }

  for(i = 0; i < sh->plan.num_accesses; i++) {
    if(sh->vq) {
      int tile_start[LF_MAXPARAMS];
      int e, bits;

      /* switch to correct [u,v,s,t] subcube of indices into tile array... */
      index = lfVQIndexTileStart(params->num, sh->vq->pbits, sh->plan.p_indices[i], sh->psamples, tile_start);
      slab->lfs[i] = &((short *)slab->lightfield)[index *
						 ((sh->nu * sh->nv * sh->ns * sh->nt) >>
						  (sh->vq->ubits + sh->vq->vbits +
						   sh->vq->sbits + sh->vq->tbits))];
      /* ... and calculate offset into the tile */
      slab->vqtiles_p[i] = sh->vq->codebook;
      bits = sh->vq->ubits+sh->vq->vbits+sh->vq->sbits+sh->vq->tbits;
      /*      for(e = params->num - 1; e >= 0; e--) {*/
      for(e = 0; e < params->num; e++) {
	slab->vqtiles_p[i] += (sh->plan.p_indices[i][e] - tile_start[e]) << bits;
	bits += sh->vq->pbits[e];
      }
    } else {
      /* switch to correct [u,v,s,t] subcube of LFPixelRGBA8 */
      index = lfIndex(params->num, sh->plan.p_indices[i], sh->psamples);
      slab->lfs[i] = &((LFPixelRGBA8 *)slab->lightfield)[index * sh->nu * sh->nv * sh->ns * sh->nt];
    }
  }
}


static void
setDrawingParams(LFOps *ops, LFInternOp *op)
{
    LFView *view = &__LF_GET_CONTEXT()->view;
    float *descr = op->op_descr;
    int i;

    for(i = 0; i < LF_MAXPARAMS; i++) {
      view->interpolate[i] = FALSE;
    }

    while (*descr != LF_NULL) {
	int param_type = (int)*descr++;
	switch(param_type) {
	case LF_DRAW_MAG_FILTER:
	    switch((int)*descr++) {
	    case LF_NEAREST_UVST:
		view->filter = 0;
		break;

	    case LF_NEAREST_UV_BILINEAR_ST:
		view->filter = LF_LERP_ST;
		break;

	    case LF_BILINEAR_UV_NEAREST_ST:
		view->filter = LF_LERP_UV;
		break;

	    case LF_QUADLINEAR_UVST:
		view->filter = LF_LERP_UV | LF_LERP_ST;
		break;
	    }
	    break;

	case LF_INTERPOLATE:
	  i = *descr++;
	  if(i >= 0 && i < LF_MAXPARAMS) {
	    view->interpolate[i] = TRUE;
	  }
	  break;
	}
    }

    for ( i=0 ; i<ops->op_slab_cnt ; i++ ) {
	LFSlab *slab = ops->op_slabs[i];
	LFShared *shared = slab->shared;

	if  (!(shared->flag & LF_SLAB_VALID_AUX)) {
	  shared->flag |= LF_SLAB_VALID_AUX;
	  lfLinearizeVQCodebook(shared->vq);
	  lfMakeIndices(shared);
	  lfPadData(slab, 0xff, 4,
#ifdef ABGR_LF	    
		    LF_PAD_FIRST
#else
		    LF_PAD_LAST
#endif	    
		    );
	  shared->u_frac_bits = 6;
	  shared->v_frac_bits = 6;
	  shared->s_frac_bits = 6;
	  shared->t_frac_bits = 6;
	}
	setInterpolationParams(slab);
    }
}

extern float pixelzoom;
void *
lfBeginDrawFB(LFOps *ops, LFInternOp *op)
{
    LFParams *params = __LF_GET_PARAMS();
    LFDriver *driver = __LF_GET_DRIVER();
    LFRenderer *renderer = __LF_GET_RENDERER();
    LFView *view = &__LF_GET_CONTEXT()->view;
    int x, y, width, height, i;
    bool_t view_changed = FALSE,
      params_changed = FALSE,
      options_changed = FALSE,
      method_changed = FALSE;
    static float old_pixelzoom;

#ifdef	STATS
    lfStartTimer();
#endif

    /* get matrices and viewport */
    (*driver->get_transform)(view->mat_model, view->mat_proj,
	&x, &y, &width, &height);
    setDrawingParams(ops, op);

    if(memcmp(view->mat_model, view->old_mat_model, sizeof(view->mat_model)) ||
       memcmp(view->mat_proj, view->old_mat_proj, sizeof(view->mat_proj)) ||
       memcmp(&view->viewport, &view->old_viewport, sizeof(view->viewport))) {
      view_changed = TRUE;
      memcpy(view->old_mat_model, view->mat_model, sizeof(view->mat_model));
      memcpy(view->old_mat_proj, view->mat_proj, sizeof(view->mat_proj));
      memcpy(&view->old_viewport, &view->viewport, sizeof(view->viewport));
    }
    for(i = 0, params_changed = FALSE, options_changed = FALSE; i < params->num; i++) {
      if(params->params[i].current != view->old_p[i]) {
	params_changed = TRUE;
	view->old_p[i] = params->params[i].current;
      }
      if(view->interpolate[i] != view->old_interpolate[i]) {
	options_changed = TRUE;
	view->old_interpolate[i] = view->interpolate[i];
      }
    }
    if(view->filter != view->old_filter) {
      view->old_filter = view->filter;
      options_changed = TRUE;
    }
    if(renderer->quality != view->old_quality) {
      view->old_quality = renderer->quality;
      options_changed = TRUE;
    }
    if(pixelzoom != old_pixelzoom) {
      old_pixelzoom = pixelzoom;
      options_changed = TRUE;
    }
    if(view->force_rendering != view->old_force_rendering) {
      view->old_force_rendering = view->force_rendering;
      method_changed = TRUE;
    }

#if 0
    /* fast interpolation is possible if only one parameter changed */
    view->fast_interp = TRUE;
    for(i = 0, view->changed_p = -1; i < params->num; i++) {
      if(params->params[i].current != view->old_p[i]) {
	if(view->changed_p == -1) {
	  view->changed_p = i;
	} else {
	  view->fast_interp = FALSE;
	}
      }
    }

    if(width == view->old_width &&
       height == view->old_height &&
       view->filter == view->old_filter &&
       !memcmp(view->mat_model, view->old_mat_model, sizeof(view->mat_model)) &&
       !memcmp(view->mat_project, view->old_mat_project, sizeof(view->mat_project)) &&
       !memcmp(view->viewport, view->old_viewport, sizeof(view->viewport))) {
    } 
#endif

    if(view_changed || options_changed || params_changed || method_changed) {
      view->rendering_done = FALSE;
      configViewport(x, y, width, height, view);

      /* clear the viewport pixel buffer */
      bzero(view->pixel_buffer, view->pixel_buffer_size);

      op->op_func = lfDrawFB;
      op->op_end  = lfEndDrawFB;
      op->op_update = lfUpdateNoop;
    } else {
      view->rendering_done = TRUE;
    }

    return NULL;
}

void
lfDrawFB(LFSlab **slabs, int nslab, const int *pos,
	const float *descr, const void *input, int input_size,
	void *output, int output_size, void *aux_ptr)
{
  LFView *view = &__LF_GET_CONTEXT()->view;

#ifdef DEBUG
  lfOutput("fov %f, obs %f/%f/%f, center %f/%f/%f, rot %f/%f/%f\n",
	   view->fov,
	   view->xobs, view->yobs, view->zobs,
	   view->xcenter, view->ycenter, view->zcenter, 
	   view->xrot, view->yrot, view->zrot); 
#endif

  if(!view->rendering_done) {
    if(nslab && !view->force_rendering) {
      lfDrawSlab(slabs[pos[LF_SLAB_POS]]);
    } else {
      /* use renderer */
      LFRenderer *renderer = &__LF_GET_CONTEXT()->field->renderer;
      
      lfRDRender(renderer, view->mat_model, view->mat_proj,
		 view->width, view->height, view->format, view->pixel_buffer);
    }
    view->rendering_done = TRUE;
  }
}

void
lfEndDrawFB(LFOps *ops, LFInternOp *op, void *aux_ptr)
{
    LFDriver *driver = __LF_GET_DRIVER();
    LFView *view = &__LF_GET_CONTEXT()->view;

    /* write the resulting looked up stuff into the framebuffer */
    (*driver->write_pixrect)(view->x, view->y, view->width, view->height,
	view->format, view->type, view->pixel_buffer);

    /* calculate intensity, if asked to do so */
    if(lfIntensity) {
      *lfIntensity = lfCalcIntensity(view->pixel_buffer, view->width, view->height,
				     view->format == LF_RGBA ? 4 : 3);
    }

#ifdef	STATS
    lfEndTimer();
#endif
}
