/* lerp.h:
 * Macros for doing linear interpolation operations.
 */

static	v0_temp, v2_temp;

#ifdef INTERP_ONLY

#define R(x) 0
#define G(x) 0
#define B(x) 0

#else

#define R(x) (x)->r
#define G(x) (x)->g
#define B(x) (x)->b

#endif

#define LERP(p0,p1,f,fbits)\
        ((int)p0 + ((f * ((int)p1-(int)p0)) >> fbits))

#define RLERP(p0,p1,f,fbits)\
        ((int)p0 + ((f * ((int)p1-(int)p0) + (1<<(fbits-1))) >> fbits))

#define BILERP(v0,v1,v2,v3,fx,fy,fxbits,fybits)\
	(v0_temp = LERP(v0,v1,fx,fxbits),\
	 v2_temp = LERP(v2,v3,fx,fxbits),\
	 LERP(v0_temp,v2_temp,fy,fybits))\

#define BILERPPIXELS(p00,p10,p01,p11,fx,fy,fx_bits,fy_bits,c) \
	(c).r = BILERP(R(p00),R(p10),R(p01),R(p11),fx,fy,fx_bits,fy_bits);\
	(c).g = BILERP(G(p00),G(p10),G(p01),G(p11),fx,fy,fx_bits,fy_bits);\
	(c).b = BILERP(B(p00),B(p10),B(p01),B(p11),fx,fy,fx_bits,fy_bits)

#define LERPPIXELS(p0,p1,f,fbits,c)\
        (c).r = LERP(R(p0),R(p1),f,fbits);\
	(c).g = LERP(G(p0),G(p1),f,fbits);\
	(c).b = LERP(B(p0),B(p1),f,fbits)


/* stuff for uncompressed lightfield */
#ifdef USE_FUNCTIONS
static LFPixelRGBA8 *LIGHTPTR_(LFPixelRGBA8 *lf, int u, int v, int s, int t,
			       int *uindex, int *vindex, int *sindex, int *tindex) {
  int index;
  static LFPixelRGBA8 pixel;

  index = uindex[u];
  index += vindex[v];
  index += sindex[s];
  index += tindex[t];
  pixel = lf[index];

  return &pixel;
}
#define	LIGHTPTR(lf,u,v,s,t) LIGHTPTR_(lf,u,v,s,t,uindex,vindex,sindex,tindex)
#else
#define	LIGHTPTR(lf,u,v,s,t)\
	(lf + uindex[u]+vindex[v]+sindex[s]+tindex[t])
#endif

#define QUADLERPRGB(lf,u,v,s,t,fu,fv,fs,ft,imgp)\
	p00 = LIGHTPTR(lf,u,v,s,t);\
	p10 = LIGHTPTR(lf,u,v,s+1,t);\
	p01 = LIGHTPTR(lf,u,v,s,t+1);\
	p11 = LIGHTPTR(lf,u,v,s+1,t+1);\
	BILERPPIXELS(p00,p10,p01,p11,fs,ft,frac_sbits,frac_tbits,c00);\
	\
	p00 = LIGHTPTR(lf,u+1,v,s,t);\
	p10 = LIGHTPTR(lf,u+1,v,s+1,t);\
	p01 = LIGHTPTR(lf,u+1,v,s,t+1);\
	p11 = LIGHTPTR(lf,u+1,v,s+1,t+1);\
	BILERPPIXELS(p00,p10,p01,p11,fs,ft,frac_sbits,frac_tbits,c10);\
	\
	p00 = LIGHTPTR(lf,u,v+1,s,t);\
	p10 = LIGHTPTR(lf,u,v+1,s+1,t);\
	p01 = LIGHTPTR(lf,u,v+1,s,t+1);\
	p11 = LIGHTPTR(lf,u,v+1,s+1,t+1);\
	BILERPPIXELS(p00,p10,p01,p11,fs,ft,frac_sbits,frac_tbits,c01);\
	\
	p00 = LIGHTPTR(lf,u+1,v+1,s,t);\
	p10 = LIGHTPTR(lf,u+1,v+1,s+1,t);\
	p01 = LIGHTPTR(lf,u+1,v+1,s,t+1);\
	p11 = LIGHTPTR(lf,u+1,v+1,s+1,t+1);\
	BILERPPIXELS(p00,p10,p01,p11,fs,ft,frac_sbits,frac_tbits,c11);\
	\
	BILERPPIXELS(&c00,&c10,&c01,&c11,fu,fv,frac_ubits,frac_vbits,(imgp))

/* Used by fast version of code that performs uv bilerping only */
#define BILERPUVRGB(lf,u,v,s,t,fu,fv,imgp)\
	p00 = LIGHTPTR(lf,u,v,s,t);\
	p10 = LIGHTPTR(lf,u+1,v,s,t);\
	p01 = LIGHTPTR(lf,u,v+1,s,t);\
	p11 = LIGHTPTR(lf,u+1,v+1,s,t);\
	BILERPPIXELS(p00,p10,p01,p11,fu,fv,frac_ubits,frac_vbits,(imgp))

/* Used by fast version of code that performs st bilerping only */
#define BILERPSTRGB(lf,u,v,s,t,fs,ft,imgp)\
	p00 = LIGHTPTR(lf,u,v,s,t);\
	p10 = LIGHTPTR(lf,u,v,s+1,t);\
	p01 = LIGHTPTR(lf,u,v,s,t+1);\
	p11 = LIGHTPTR(lf,u,v,s+1,t+1);\
	BILERPPIXELS(p00,p10,p01,p11,fs,ft,frac_sbits,frac_tbits,(imgp))

/* Used by fast version of code that performs point sampling only */
#if defined(ABGR_LF) == defined(ABGR_OUTPUT)
#define POINTSAMPLERGB(lf,u,v,s,t,imgp) \
        *((LFInt32 *)&(imgp)) = *((LFInt32 *)LIGHTPTR(lf,u,v,s,t))
#else
#define POINTSAMPLERGB(lf,u,v,s,t,imgp)\
	p00 = LIGHTPTR(lf,u,v,s,t);\
	(imgp).r = p00->r;\
	(imgp).g = p00->g;\
	(imgp).b = p00->b
#endif

/* stuff for vq */
#ifdef USE_FUNCTIONS
static LFPixelRGBA8 *VQLIGHTPTR_(short *lf, LFPixelRGBA8 *vqtile_p,
				 int u, int v, int s, int t, int tile_shift,
				 int *uindex, int *vindex, int *sindex, int *tindex,
				 int *vqtile_uindex, int *vqtile_vindex,
				 int *vqtile_sindex, int *vqtile_tindex) { 
  int index, vqindex;
  static LFPixelRGBA8 pixel;

  index = uindex[u];
  index += vindex[v];
  index += sindex[s];
  index += tindex[t];
  vqindex = lf[index]<<tile_shift;
  vqindex += vqtile_uindex[u];
  vqindex += vqtile_vindex[v];
  vqindex += vqtile_sindex[s];
  vqindex += vqtile_tindex[t];
  pixel = vqtile_p[vqindex];

  return &pixel;
}
#define VQLIGHTPTR(lf,vqtile_p,u,v,s,t) VQLIGHTPTR_(lf,vqtile_p,u,v,s,t,tile_shift,uindex,vindex,sindex,tindex,vqtile_uindex,vqtile_vindex,vqtile_sindex,vqtile_tindex)
#else
#define	VQLIGHTPTR(lf,vqtile_p,u,v,s,t)\
	(vqtile_p + \
	 (lf[uindex[u]+vindex[v]+sindex[s]+tindex[t]]<<tile_shift) + \
	 vqtile_uindex[u]+vqtile_vindex[v]+vqtile_sindex[s]+vqtile_tindex[t])
#endif

#define QUADLERPVQRGB(lf,vqtile_p,u,v,s,t,fu,fv,fs,ft,imgp)\
	p00 = VQLIGHTPTR(lf,vqtile_p,u,v,s,t);\
	p10 = VQLIGHTPTR(lf,vqtile_p,u,v,s+1,t);\
	p01 = VQLIGHTPTR(lf,vqtile_p,u,v,s,t+1);\
	p11 = VQLIGHTPTR(lf,vqtile_p,u,v,s+1,t+1);\
	BILERPPIXELS(p00,p10,p01,p11,fs,ft,frac_sbits,frac_tbits,c00);\
	\
	p00 = VQLIGHTPTR(lf,vqtile_p,u+1,v,s,t);\
	p10 = VQLIGHTPTR(lf,vqtile_p,u+1,v,s+1,t);\
	p01 = VQLIGHTPTR(lf,vqtile_p,u+1,v,s,t+1);\
	p11 = VQLIGHTPTR(lf,vqtile_p,u+1,v,s+1,t+1);\
	BILERPPIXELS(p00,p10,p01,p11,fs,ft,frac_sbits,frac_tbits,c10);\
	\
	p00 = VQLIGHTPTR(lf,vqtile_p,u,v+1,s,t);\
	p10 = VQLIGHTPTR(lf,vqtile_p,u,v+1,s+1,t);\
	p01 = VQLIGHTPTR(lf,vqtile_p,u,v+1,s,t+1);\
	p11 = VQLIGHTPTR(lf,vqtile_p,u,v+1,s+1,t+1);\
	BILERPPIXELS(p00,p10,p01,p11,fs,ft,frac_sbits,frac_tbits,c01);\
	\
	p00 = VQLIGHTPTR(lf,vqtile_p,u+1,v+1,s,t);\
	p10 = VQLIGHTPTR(lf,vqtile_p,u+1,v+1,s+1,t);\
	p01 = VQLIGHTPTR(lf,vqtile_p,u+1,v+1,s,t+1);\
	p11 = VQLIGHTPTR(lf,vqtile_p,u+1,v+1,s+1,t+1);\
	BILERPPIXELS(p00,p10,p01,p11,fs,ft,frac_sbits,frac_tbits,c11);\
	\
	BILERPPIXELS(&c00,&c10,&c01,&c11,fu,fv,frac_ubits,frac_vbits,(imgp))

/* Used by fast version of code that performs uv bilerping only */
#define BILERPUVVQRGB(lf,vqtile_p,u,v,s,t,fu,fv,imgp)\
	p00 = VQLIGHTPTR(lf,vqtile_p,u,v,s,t);\
	p10 = VQLIGHTPTR(lf,vqtile_p,u+1,v,s,t);\
	p01 = VQLIGHTPTR(lf,vqtile_p,u,v+1,s,t);\
	p11 = VQLIGHTPTR(lf,vqtile_p,u+1,v+1,s,t);\
	BILERPPIXELS(p00,p10,p01,p11,fu,fv,frac_ubits,frac_vbits,(imgp))

/* Used by fast version of code that performs st bilerping only */
#define BILERPSTVQRGB(lf,vqtile_p,u,v,s,t,fs,ft,imgp)\
	p00 = VQLIGHTPTR(lf,vqtile_p,u,v,s,t);\
	p10 = VQLIGHTPTR(lf,vqtile_p,u,v,s+1,t);\
	p01 = VQLIGHTPTR(lf,vqtile_p,u,v,s,t+1);\
	p11 = VQLIGHTPTR(lf,vqtile_p,u,v,s+1,t+1);\
	BILERPPIXELS(p00,p10,p01,p11,fs,ft,frac_sbits,frac_tbits,(imgp))

/* Used by fast version of code that performs point sampling only */
#if defined(ABGR_LF) == defined(ABGR_OUTPUT) && defined(ALPHA_OUT)
#define POINTSAMPLEVQRGB(lf,vqtile_p,u,v,s,t,imgp) \
        *((LFInt32 *)&(imgp)) = *((LFInt32 *)VQLIGHTPTR(lf,vqtile_p,u,v,s,t))
#else
#define POINTSAMPLEVQRGB(lf,vqtile_p,u,v,s,t,imgp)\
	p00 = VQLIGHTPTR(lf,vqtile_p,u,v,s,t);\
	(imgp).r = p00->r;\
	(imgp).g = p00->g;\
	(imgp).b = p00->b
#endif
