/*
 * opengl_driver.c
 *
 * driver functions for OpenGL
 */
#include <GL/gl.h>
#include "lightfield.h"

/* Pixelzoom allows the final image to be scaled up, so that it
 * is bigger (more visible) on the screen.
 */
float pixelzoom = 2.0;

/* function to get current transformation information */
void
opengl_get_transform(Matrix4 mat_model, Matrix4 mat_proj,
	int *x, int *y, int *width, int *height)
{
    GLint v[4];
    int i;

    glGetFloatv(GL_MODELVIEW_MATRIX, (float *)mat_model);
    TRANSMAT4(mat_model, mat_model);

    glGetFloatv(GL_PROJECTION_MATRIX, (float *)mat_proj);
    TRANSMAT4(mat_proj, mat_proj);

    glGetIntegerv(GL_VIEWPORT, v);
    /* Adjust size for pixel scaling. */
    for(i = 0; i < 4; i++)
      v[i] /= pixelzoom;
    *x = v[0];
    *y = v[1];
    *width  = v[2];
    *height = v[3];
}




/* function to draw stuff into current framebuffer */
void
opengl_write_pixrect(int x, int y, int w, int h,
	int format, int type, void *pixels)
{
  #ifdef ABGR_OUTPUT
  unsigned char *rgbmem = pixels;
  int i, j;
  #endif

  glPixelZoom(pixelzoom, pixelzoom);
  #ifdef ALPHA_OUT
  #ifdef ABGR_OUTPUT
  for (i=0; i < h; i++)
    for (j=0; j < w; j++) {	/* convert from ABGR to RGBA */
      unsigned char tmp;
      
      tmp = rgbmem[0];
      rgbmem[0] = rgbmem[3];
      rgbmem[3] = tmp;
      tmp = rgbmem[1];
      rgbmem[1] = rgbmem[2];
      rgbmem[2] = tmp;
      rgbmem += 4;
    }
  #endif
  #else
  #ifdef ABGR_OUTPUT
  for (i=0; i < h; i++)
    for (j=0; j < w; j++) {	/* convert from BGR to RGB */
      unsigned char tmp;
      
      tmp = rgbmem[0];
      rgbmem[0] = rgbmem[2];
      rgbmem[2] = tmp;
      rgbmem += 3;
    }
  #endif
  #endif
  glDrawPixels(w, h,
	       #ifdef ALPHA_OUT
	       GL_RGBA,
	       #else
	       GL_RGB,
	       #endif
	       GL_UNSIGNED_BYTE, pixels);
}

LFDriver opengl_driver = {
    opengl_get_transform,
    opengl_write_pixrect,
};

LFDriver *__lf_driver = &opengl_driver;


