/*
** renderer.c - interface to builtin renderers
*/

#include "renderer.h"

bool_t lfRDAlloc(LFRenderer *renderer)
{
  switch(renderer->type) {
  case LF_RENDERER_VOLPACK:
    return(lfVPAlloc(renderer));
    break;
  case LF_RENDERER_RENDERMAN:
    return(lfRIBAlloc(renderer));
    break;
  case LF_RENDERER_VOLVIS:
    return(lfVVAlloc(renderer));
    break;
  default:
    return(TRUE);
  }
}
void lfRDDone(LFRenderer *renderer)
{
  switch(renderer->type) {
  case LF_RENDERER_VOLPACK:
    lfVPDone(renderer);
    break;
  case LF_RENDERER_RENDERMAN:
    lfRIBDone(renderer);
    break;
  case LF_RENDERER_VOLVIS:
    lfVVDone(renderer);
    break;
  default:
    return;
  }
}
bool_t lfRDRender(LFRenderer *renderer,
		   Matrix4 mat_model, Matrix4 mat_proj, int width, int height,
		   int format, void *pixels)
{
  if(renderer->flags & LFRF_NO_RENDERING) {
    lfError("lfRDRender: renderer had been disabled!\n");
    return FALSE;
  }
  switch(renderer->type) {
  case LF_RENDERER_VOLPACK:
    return(lfVPRender(renderer, mat_model, mat_proj, width, height, format, pixels));
    break;
  case LF_RENDERER_RENDERMAN:
    return(lfRIBRender(renderer, mat_model, mat_proj, width, height, format, pixels));
    break;
  case LF_RENDERER_VOLVIS:
    return(lfVVRender(renderer, mat_model, mat_proj, width, height, format, pixels));
    break;
  case LF_RENDERER_DUMMY:
  default:
    {
      LFParams *params = &__LF_GET_CONTEXT()->field->params;
      unsigned char r, g, b;
      
      if(params->num > 2) {
	b = params->params[2].current;
      } else {
	b = 128;
      }
      if(params->num > 1) {
	g = params->params[1].current;
      } else {
	g = 64;
      }
      if(params->num > 0) {
	r = params->params[0].current;
      } else {
	r = 32;
      }
      
      switch(format) {
      case LF_RGBA:
	{
	  LFPixelRGBA8 *dst = pixels;
	  while(--height >= 0) {
	    int tmp;
	    
	    for(tmp = width; --tmp >= 0; dst++) {
	      dst->a = 0;
	      dst->r = r;
	      dst->g = g;
	      dst->b = b;
	    }
	  }
	  break;
	}
      case LF_RGB:
	{
	  LFPixelRGB8 *dst = pixels;
	  while(--height >= 0) {
	    int tmp;
	    
	    for(tmp = width; --tmp >= 0; dst++) {
	      dst->r = r;
	      dst->g = g;
	      dst->b = b;
	    }
	  }
	  break;
	}
      default:
	lfError("lfRDRender(): unknown format %d\n", format);
	exit(1);
	break;
      }
    }
    return TRUE;
    break;
  }
  return FALSE;
}

void lfRDRenderParams(LFRenderer *renderer)
{
  switch(renderer->type) {
  case LF_RENDERER_VOLPACK:
    lfVPRenderParams(renderer);
    break;
  case LF_RENDERER_VOLVIS:
    lfVVRenderParams(renderer);
    break;
  default:
    return;
  }
}
