/*
** interface to VolVis library
*/

#ifndef NO_VV

#include "C_volvis.h"
#include "C_fio.h"
#include "C_render.h"
#include "renderer.h"
#include "volume.h"

#include <GL/gl.h>

#include <stdlib.h>
#include <math.h>
#include <strings.h>
#ifndef min
#define min(x,y) ((x) < (y) ? (x) : (y))
#endif

/* XXX: the following variable should better be put into
   a context for this renderer... */
C_World world;       /* the VolVis world */
enum ParamMode
{
  ISO_SURFACE,
  OPACITY,
  PM_MAX
} param_mode;
int Volume[PM_MAX + 1]; /* the first volumes with an ISO surface and opacity ray casting */
#define RayCastVol Volume[OPACITY]
#define ISOSurfaceVol Volume[ISO_SURFACE]
#define DefaultVol    Volume[PM_MAX]
vpContext *vpc;      /* for viewing calculations with lfVolSetPosition()
			and fast preview */
bool_t vpcVolumeValid; /* vpc can be used for rendering */
static int old_vpc_position, old_vpc_steepness;
static bool_t VPCVolumeValid(void)
{
  if(vpcVolumeValid) {
    return TRUE;
  }

  if(RayCastVol != -1) {
    if(lfVPInitVPC(vpc,
		   world.volume[RayCastVol]->data.scalar_data->scalar_data,
		   world.volume[RayCastVol]->data.scalar_data->x_size_voxels,
		   world.volume[RayCastVol]->data.scalar_data->y_size_voxels,
		   world.volume[RayCastVol]->data.scalar_data->z_size_voxels, 0)) {
      vpcVolumeValid = TRUE;
      old_vpc_position = -1, old_vpc_steepness = -1;
    }
  }

  return vpcVolumeValid;
}




/* required by PARC rendering, which is not enabled at the moment */
C_NavInfo nav_info;
float C_temp_float1;
#ifdef OGL_RENDERIN
void C_parc_polygon_projection(C_NavProjectionWinInfo *pj_info)
{
  /* dummy for src/[O]GL_SRC/GL_nav_projection.c */
  lfError("C_parc_polygon_projection: dummy function called!\n");
}
#else
/* dummies for src/MOTIF_SRC/C_ren_approx_marching_cubes.c */
void C_xgrph_draw_shaded_tri(void)
{
  lfError("C_xgrph_draw_shaded_tri: dummy function called!\n");
}
void C_xgrph_draw_wire_tri(void)
{
  lfError("C_xgrph_draw_wire_tri: dummy function called!\n");
}
#endif

void C_initialize_world(C_World *world)
{
  bzero(world, sizeof(*world));

  /* Initialize World Coordinate System */
  C_Set_Orig( world->orig_c_sys, 0.0, 0.0, 0.0 );
  C_Set_Axes( world->orig_c_sys, 
	      1.0, 0.0, 0.0, 
	      0.0, 1.0, 0.0, 
	      0.0, 0.0, 1.0 );

  /* init rendering */
  initialize_MC_edge_table();
}



/*
** set new transfer parameters and return TRUE if they have changed
*/
static int old_position = -1, old_steepness = -1;
static bool_t setTransfer(LFRenderer *renderer)
{
  LFParams *params = &__LF_GET_CONTEXT()->field->params;
  int position = params->params[renderer->arg_indices[0]].current;
  int steepness = params->params[renderer->arg_indices[1]].current;
  int orig_stp = steepness, orig_pos = position;

  if(position == old_position &&
     steepness == old_steepness)
    return TRUE;

  lfVolSetRamp(renderer);

  switch(param_mode) {
  case OPACITY:
    /* set new opacity transfer */
    memcpy(world.volume[RayCastVol]->data.scalar_data->seg.s_opacity, density_ramp,
	   sizeof(*world.volume[RayCastVol]->data.scalar_data->seg.s_opacity));
    break;
  case ISO_SURFACE:
    /* set new iso value */
    world.volume[ISOSurfaceVol]->data.scalar_data->seg.isovalue = position;
    break;
  default:
    /* cannot change anything */
    break;
  }

  old_position = orig_pos;
  old_steepness = orig_stp;
  return TRUE;
}


bool_t lfVVAlloc(LFRenderer *renderer)
{
  char fnm[256];
  int i;
  float scale, trans_x, trans_y, trans_z;
  bool_t force_iso = FALSE,
    force_opacity = FALSE;
  char *dir = renderer->config;

  /* set parameter rendering and return immediately, if requested */
  renderer->aspect = 1.5;
  if(renderer->flags & LFRF_NO_RENDERING) {
    return TRUE;
  }
  renderer->sup_qualities = (1<<4)-1;

  if(vpc) {
    vpDestroyContext(vpc);
  }
  vpc = vpCreateContext();
  vpcVolumeValid = FALSE;
  if(!vpc) {
    lfError("lfVVAlloc: no vp context\n");
    return FALSE;
  }

  if(!strncmp(dir, "ISO", 3)) {
    force_iso = TRUE;
    dir += 3;
  } else if(!strncmp(dir, "OPACITY", 7)) {
    force_opacity = TRUE;
    dir += 7;
  }

  C_initialize_world(&world);
  old_position = -1, old_steepness = -1;

  if(dir[0] == '/' ||
     renderer->lidDir[0] == 0) {
    strcpy(fnm, dir);
  } else {
    sprintf(fnm, "%s/%s", renderer->lidDir, dir);
  }
  lfOutput("lfVVAlloc: loading %s... ", fnm);
  if(C_read_env(&world, fnm)) {
    lfError("cannot read VolVis environment file %s\n", fnm);
    return FALSE;
  }

  /* search for volumes to work on */
  ISOSurfaceVol = -1;
  RayCastVol = -1;
  DefaultVol = -1;
  for(i = 0; i < world.num_volumes; i++) {
    if(world.volume[i]->data_type == C_RECT_SCALAR_DATA) {
      if(world.volume[i]->data.scalar_data->seg_type == C_SCALAR_OPACITY) {
	if(RayCastVol == -1) {
	  RayCastVol = i;
	  lfOutput("opacity transfer: volume %s\n", world.volume[i]->volume_name);
	} else {
	  lfOutput("Leaving opacity transfer of volume %s unchanged.\n", world.volume[i]->volume_name);
	}
      } else if(world.volume[i]->data.scalar_data->seg_type == C_ISO_SURFACE) {
	if(ISOSurfaceVol == -1) {
	  ISOSurfaceVol = i;
	  lfOutput("iso surface: volume %s\n", world.volume[i]->volume_name);
	} else {
	  lfOutput("Leave iso value of volume %s unchanged.\n", world.volume[i]->volume_name);
	}
      } else {
	if(DefaultVol == -1) {
	  DefaultVol = i;
	}
	lfOutput("Leaving volume %s with segmentation type %d unchanged.\n",
		 world.volume[i]->volume_name, world.volume[i]->data.scalar_data->seg_type);
      }
    } else {
      if(DefaultVol == -1) {
	DefaultVol = i;
      }
      lfOutput("Leaving volume %s with data type %d unchanged.\n",
	       world.volume[i]->volume_name, world.volume[i]->data_type);
    }
  }

  /* switch some volumes? */
  if(force_iso) {
    if(ISOSurfaceVol == -1) {
      if(RayCastVol != -1) {
	world.volume[RayCastVol]->data.scalar_data->seg_type = C_ISO_SURFACE;
	lfOutput("Render volume %s with ISO surface instead of opacity .\n",
		 world.volume[RayCastVol]->volume_name);
	ISOSurfaceVol = RayCastVol;
	RayCastVol = -1;
      }
    }
  } else if(force_opacity) {
    if(RayCastVol == -1) {
      if(ISOSurfaceVol != -1) {
	world.volume[ISOSurfaceVol]->data.scalar_data->seg_type = C_SCALAR_OPACITY;
	world.volume[ISOSurfaceVol]->data.scalar_data->seg.s_opacity =
	  malloc(sizeof(*world.volume[ISOSurfaceVol]->data.scalar_data->seg.s_opacity));
	lfOutput("Render volume %s with opacity instead of ISO surface.\n",
		 world.volume[ISOSurfaceVol]->volume_name);
	RayCastVol = ISOSurfaceVol;
	ISOSurfaceVol = -1;
      }
    }
  }

  if(RayCastVol != -1) {
    param_mode = OPACITY;
    lfOutput("Do opacity transfer with parameters.\n");
  } else if(ISOSurfaceVol != -1) {
    param_mode = ISO_SURFACE;
    lfOutput("Set iso surface with parameter(s).\n");
  } else {
    param_mode = PM_MAX;
    lfOutput("Cannot change any volume.\n");
  }

  /* adjust world and object coordinates to lightfield model (unit cube) */
  if(Volume[param_mode] != -1) {
    trans_x = -world.volume[Volume[param_mode]]->x_size_units/2;
    trans_y = -world.volume[Volume[param_mode]]->y_size_units/2;
    trans_z = -world.volume[Volume[param_mode]]->z_size_units/2;
  } else {
    trans_x =
      trans_y =
      trans_z = 0;
  }
  #define SQUARE(x) ((x)*(x))
  scale = 1.96 / sqrt(SQUARE(world.ren_view.orig_c_sys.origin.x + trans_x) +
		      SQUARE(world.ren_view.orig_c_sys.origin.y + trans_y) +
		      SQUARE(world.ren_view.orig_c_sys.origin.z + trans_z));
  for(i = 0; i < world.num_volumes; i++) {
    C_translate_volume_local(world.volume[i],
			     trans_x,
			     trans_y,
			     trans_z);
    world.volume[i]->x_scale_factor *= scale;
    world.volume[i]->y_scale_factor *= scale;
    world.volume[i]->z_scale_factor *= scale;
  }
  world.ren_view.orig_c_sys.origin.x = 0;
  world.ren_view.orig_c_sys.origin.y = 0;
  world.ren_view.orig_c_sys.origin.z = 0;
  /* C_translate_view_local(&world.ren_view, trans_x, trans_y, trans_z); */
  world.ren_view.transform.matrix[3][3] /= scale;

  /* calculate histogram */
  if((i = RayCastVol) != -1 ||
     (i = ISOSurfaceVol) != -1) {
    volCalcHistogram(world.volume[i]->data.scalar_data->scalar_data,
		     world.volume[i]->data.scalar_data->x_size_voxels *
		     world.volume[i]->data.scalar_data->y_size_voxels *
		     world.volume[i]->data.scalar_data->z_size_voxels,
		     volHistogram, &volHisMax);
  }

  lfOutput("done.\n");
  return TRUE;
}

void lfVVDone(LFRenderer *renderer)
{
  /* todo: clean up world */
}

bool_t lfVVRender(LFRenderer *renderer,
		   Matrix4 mat_model, Matrix4 mat_proj, int width, int height,
		   int format, void *pixels)
{
  LFView *view = &__LF_GET_CONTEXT()->view;
  C_Image image;
  int num_pixels = width * height;
  bool_t result = FALSE;

  if(renderer->quality == 0 &&
     param_mode == OPACITY) {
    /* do VolPack rendering instead */
    return(VPCVolumeValid() &&
	   lfVPRenderVPC(vpc, mat_model, mat_proj, width, height,
			 format, pixels, NULL));
  }

  bzero(&image, sizeof(image));
  if((image.red = calloc(num_pixels * 3, 1))) {
    C_CoordSys old_c_sys;
    C_Matrix old_transform;
    vpMatrix4 vp_view;
    C_Matrix vv_view;
    int i, j;

    old_c_sys = world.ren_view.c_sys;
    old_transform = world.ren_view.transform;

    image.green = &image.red[num_pixels];
    image.blue  = &image.green[num_pixels];
    world.ren_view.width_pixels =
      image.width = width;
    world.ren_view.height_pixels =
      image.height = height;
    image.depth = 3;

    world.ren_view.stereo_project = 0;
    switch(renderer->quality) {
    case 0:
      world.ren_view.rt_image_accuracy = C_CRUDE_IMAGE_ACCURACY;
      break;
    case 1:
      world.ren_view.rt_image_accuracy = C_LOW_IMAGE_ACCURACY;
      break;
    case 2:
      world.ren_view.rt_image_accuracy = C_MEDIUM_IMAGE_ACCURACY;
      break;
    default:
      world.ren_view.rt_image_accuracy = C_HIGH_IMAGE_ACCURACY;
      break;
    }

    /* set view position by doing the VolPack calculations
       and then multiplying them onto transform and c_sys */
    lfVolSetPosition(vpc, width, height, TRUE);
    vpGetMatrix(vpc, VP_VIEW, vp_view);
    for(i = 0; i < 4; i++) {
      for(j = 0; j < 4; j++) {
	vv_view.matrix[i][j] = vp_view[i][j];
      }
    }
    /* c_sys = transform*orig_c_sys;
    ** transform = view*transform => c_sys = view*c_sys
    */
    C_matrix_multiply(C_POST_MULTIPLY, &world.ren_view.transform, &vv_view);
    world.ren_view.fov = view->fov != 0.0 ?
      view->fov * PI / 180 :
      2 * atan((view->screen_upper - view->screen_lower)/2.0);
    C_transform_c_sys(&world.ren_view.orig_c_sys, &world.ren_view.c_sys,
		      &world.ren_view.transform);

    setTransfer(renderer);

    if(!C_cast_rays(&world, &world.ren_view, &image, NULL, NULL, NULL, NULL)) {
      C_Byte *r = image.red,
	*g = image.green,
	*b = image.blue,
	*out = pixels;
      int line, line_width = width * (format == LF_RGBA ? 4 : 3);

      line = height;
      while(--line >= 0) {
	out = &((C_Byte *)pixels)[line_width * line];
	num_pixels = width;
	while(--num_pixels >= 0) {
	  *out++ = *r++;
	  *out++ = *g++;
	  *out++ = *b++;
	  if(format == LF_RGBA) {
	    *out++ = 0xFF;
	  }
	}
      }
      result = TRUE;
    } else {
      lfError("lfVVRender: rendering failed\n");
    }
    world.ren_view.c_sys = old_c_sys;
    world.ren_view.transform = old_transform;
  }

  free(image.red);
  return result;
}


void lfVVRenderParams(LFRenderer *renderer)
{
  lfVolRenderParams(renderer);
}


/*
** code required by VolVis libs
*/

void C_error_message(char *msg)
{
  lfError(msg);
}
void accelerate_MC(int vol_id)
{
}

void C_free_texture( texture )
C_Texture		*texture;
{
	if ( !texture ) return;


	switch ( texture->texture_type )
	{
	  case C_VOLUME_TEXTURE:
		if ( texture->texture.voltex->red )
			free( texture->texture.voltex->red );

		if ( texture->texture.voltex->green )
			free( texture->texture.voltex->green );

		if ( texture->texture.voltex->blue )
			free( texture->texture.voltex->blue );

		free( texture->texture.voltex );

		break;

	  default:
		C_error_message("Unknown texture type");
		break;
	}
		
}
int C_get_texture_index( texture_name )
char			*texture_name;
{
	extern C_World	world;

	int		loop;

	for (loop = 0; loop < world.num_textures; loop++)
	{
		if (strcmp( texture_name, world.texture[loop]->texture_name )
			== 0 ) 
		{
			return loop;
		}
	}
		
	return C_NO_MATCH;
}

void C_free_volume( volume )
C_Volume	*volume;
{

	void		C_free_scalar_data();
	void		C_free_geo_data();
	void		C_free_cut_geom();
	void		C_free_volume_texture();
	void		C_free_MC_approx_list();
	void		C_free_parc_info();

	int		loop;


	switch ( volume->data_type )
	{
	  case C_RECT_SCALAR_DATA:

		C_free_scalar_data( volume->data.scalar_data );
		break;
		
	  case C_GEOMETRIC_DATA:

		C_free_geo_data( volume->data.geo_data );
		break;

	  default:

		C_error_message("fix this\n");
		break;
	}

	for ( loop = 0; loop < volume->num_cuts; loop++ )
		C_free_cut_geom( volume->local_cut[loop] );

	if ( volume->color_data.single_color )
		free( volume->color_data.single_color );

	if ( volume->color_data.color_array )
		free( volume->color_data.color_array );

	if ( volume->texture )
	  volume->texture->references--;

	if ( volume->local_shade.simple_shade )
		free( volume->local_shade.simple_shade );

	if ( volume->local_shade.array_shade )
		free( volume->local_shade.array_shade );

	if ( volume->volume_name )
		free( volume->volume_name );

	if ( volume->file_name )
		free( volume->file_name );
	
	if ( volume->icon_image.red )
		free( volume->icon_image.red );
	if ( volume->icon_image.green )
		free( volume->icon_image.green );
	if ( volume->icon_image.blue )
		free( volume->icon_image.blue );
	
	C_free_MC_approx_list( volume );

	C_free_parc_info( volume->p_info );

	free( volume );

}

void C_free_scalar_data( scalar_data )
C_ScalarData	*scalar_data;
{
	if (!scalar_data) return;

	(*(scalar_data->references))--;

	if ( *(scalar_data->references) <= 0 )
	{
	    switch ( scalar_data->seg_type )
	    {
		case C_ISO_SURFACE:
		    break;

		case C_SCALAR_OPACITY:
		    if ( scalar_data->seg.s_opacity )
			free( scalar_data->seg.s_opacity );
		    break;
	    default:
	      /* several cases had been reported as missing */
	      break;
	    }

	    if ( scalar_data->scalar_data )
		free( scalar_data->scalar_data );

	    free ( scalar_data->references );

	    free( scalar_data );
	}
}

void C_free_geo_data( geo_data )
C_GeoData		*geo_data;
{
	if ( !geo_data ) return;

	switch ( geo_data->geo_type )
	{
	    case C_POLYGON:
		if ( geo_data->geo_object.polygon )
			free ( geo_data->geo_object.polygon );
		break;

	    case C_MULTIPOLY:
	    case C_HOLLOW_MULTIPOLY:
		if ( geo_data->geo_object.polygon )
			free ( geo_data->geo_object.polygon );
		break;

	    case C_TORUS:
	    case C_HOLLOW_TORUS:
		if ( geo_data->geo_object.torus )
			free ( geo_data->geo_object.torus );
		break;
	    default:
	      /* several cases had been reported as missing */
	      break;
	}

	free( geo_data );
}

void C_free_cut_geom( cut_geom )
C_CutGeom		*cut_geom;
{
	if ( !cut_geom ) return;

	switch( cut_geom->cut_type )
	{
	    case C_PLANE_CUT:
		if ( cut_geom->cut.plane )
			free( cut_geom->cut.plane );
		break;

	    case C_SIMPLE_BOX_CUT:
		if ( cut_geom->cut.box_cut )
			free( cut_geom->cut.box_cut );
		break;
	    default:
	      /* several cases had been reported as missing */
	      break;
	}

	if ( cut_geom->cut_name )
		free( cut_geom->cut_name );

	if ( cut_geom->cut_shade.simple_shade )
		free( cut_geom->cut_shade.simple_shade );

	if ( cut_geom->cut_color.single_color )
		free( cut_geom->cut_color.single_color );

	if ( cut_geom->cut_color.color_array )
		free( cut_geom->cut_color.color_array );

	free( cut_geom );
}

void C_free_parc_info( parc_info )
C_ParcInfo		*parc_info;
{
	C_ParcSubNode	*node_ptr;
	C_ParcSubNode	*next_node_ptr;

	if ( !parc_info ) return;

	node_ptr = parc_info->sub_list;
	
	while ( node_ptr )
	{
		next_node_ptr = node_ptr->next;
		free( node_ptr );
		node_ptr = next_node_ptr;
	}

	free( parc_info );
}

void C_free_light( light )
C_Light	*light;
{

	switch ( light->light_type )
	{
	  case C_POINT_LIGHT:

		if ( light->light.light_point )
			free( light->light.light_point );	
		break;
		
	  default:

		C_error_message("fix this\n");
		break;
	}

	if ( light->light_name )
		free( light->light_name );

	if ( light->file_name )
		free( light->file_name );
}

#else /* NO_VV */

#include "renderer.h"
#include "volume.h"

bool_t lfVVAlloc(LFRenderer *renderer)
{
  if(renderer->flags & LFRF_NO_RENDERING) {
    return TRUE;
  } else {
    return FALSE;
  }
}

void lfVVDone(LFRenderer *renderer) { }

bool_t lfVVRender(LFRenderer *renderer,
		   Matrix4 mat_model, Matrix4 mat_proj, int width, int height,
		   int format, void *pixels)
{ return FALSE; }

void lfVVRenderParams(LFRenderer *renderer)
{
  lfVolRenderParams(renderer);
}

#endif /* NO_VV */



