/*
 *  view.c, version 1.0alpha1.
 *    View a light field.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#ifndef M_PI
#define M_PI            3.14159265358979323846
#endif
#ifndef min
#define min(x,y) ((x) < (y) ? (x) : (y))
#endif
#ifndef max
#define max(x,y) ((x) > (y) ? (x) : (y))
#endif
#include <GL/glut.h>
#define fcos(x) cos(x)
#define ftan(x) tan(x)
#define fsin(x) sin(x)
#define fatan(x) atan(x)
#include <unistd.h>
#include "lightfield.h"

#define FZ 0.0
#define EPSILON 0.001

#define DEFAULTSIZEX 192
#define DEFAULTSIZEY 192

#define BASENAME "."

#define MENU_DUMMY      0
#define MENU_POINT	1
#define MENU_UVINTER	2
#define MENU_STINTER	3
#define MENU_ZOOM1      4
#define MENU_ZOOM2      5
#define MENU_ZOOM3      6
#define MENU_REFINE1    7
#define MENU_REFINE2    8
#define MENU_REFINE3    9
#define MENU_RESET	10
#define MENU_EXIT	11
#define MENU_INTENSITY  12
#define MENU_REFINEQ    13
#define MENU_RENDERING  14

#define MENU_QUALITY(x) (200 + (x))
#define MENU_INTERP(x)  (100 + (x))

#define QUALITY_CRUDE  "crude"
#define QUALITY_LOW    "low"
#define QUALITY_MEDIUM "medium"
#define QUALITY_HIGH   "high"

static char *quality_names[] =
{
  QUALITY_CRUDE,
  QUALITY_LOW,
  QUALITY_MEDIUM,
  QUALITY_HIGH
};
#define NUM_QUALITIES (sizeof(quality_names)/sizeof(quality_names[0]))


#define RESETXOBS 0.0
#define RESETYOBS 0.0
#define RESETZOBS 1.985

#define INITXCENTER 0.0
#define INITYCENTER 0.0
#define INITZCENTER 0.0

#define	UVLERP	0x1
#define	STLERP	0x2

#define REFINE_INTERP 0x1
#define REFINE_ZOOM   0x2
#define REFINE_QUAL   0x4

#define	ROTATEX	0x1
#define	ROTATEY	0x2
#define	ROTATEZ	0x4


#define SIGN(x) (((x)<0) ? -1 : 1)

float xobs = RESETXOBS, yobs = RESETYOBS, zobs = RESETZOBS;
float xcenter = INITXCENTER;
float ycenter = INITYCENTER;
float zcenter = INITZCENTER;
float fov = 40;
long xsize = DEFAULTSIZEX;
long ysize = DEFAULTSIZEY;
extern float pixelzoom;	   /* defined in irisgl_driver.c */

/* This is used as the temporary file if we need to gunzip it. */
char tempLif[1000] = "/tmp/current.lif";


typedef enum {
  LF_LID,
  LF_LIF,
  LF_LIFBADVER,
  LF_GZ,
  LF_UNKNOWN
} LF_FileType;

/* in degrees */
float xrot = 0;
float yrot = 0;
float zrot = 0;
#define MAX_ROT_ANGLE 10
#define ROT_ANGLE_STEPS 1

char *baseName = NULL;
int interp = UVLERP;
bool_t autoRefine = REFINE_INTERP;
int refineredraw = REFINE_INTERP;
bool_t noLight, createImages, noImages, uvFilter;
int quality = -1;
/* Four draw descriptions, for the 4 types of interpolation */
#define MAX_INTERPOLATE 5 /* allow up to 5 interpolated variables */
bool_t interp_params[MAX_INTERPOLATE];
float intensity;
float draw_descr[4][MAX_INTERPOLATE * 2 + 3] = {
  {LF_INTERPOLATE,-1,LF_INTERPOLATE,-1,LF_INTERPOLATE,-1,LF_INTERPOLATE,-1,LF_INTERPOLATE,-1,
   LF_DRAW_MAG_FILTER, LF_NEAREST_UVST, LF_NULL },
  {LF_INTERPOLATE,-1,LF_INTERPOLATE,-1,LF_INTERPOLATE,-1,LF_INTERPOLATE,-1,LF_INTERPOLATE,-1,
   LF_DRAW_MAG_FILTER, LF_BILINEAR_UV_NEAREST_ST, LF_NULL },
  {LF_INTERPOLATE,-1,LF_INTERPOLATE,-1,LF_INTERPOLATE,-1,LF_INTERPOLATE,-1,LF_INTERPOLATE,-1,
   LF_DRAW_MAG_FILTER, LF_NEAREST_UV_BILINEAR_ST, LF_NULL },
  {LF_INTERPOLATE,-1,LF_INTERPOLATE,-1,LF_INTERPOLATE,-1,LF_INTERPOLATE,-1,LF_INTERPOLATE,-1,
   LF_DRAW_MAG_FILTER, LF_QUADLINEAR_UVST, LF_NULL }};
int leftdown = 0, middledown = 0; /* status of mouse buttons */
int shift = 0;                    /* status of shift key */
float lx, ly;                     /* last mouse position as percentage of window size */
int single_frame = 0;             /* do single step in forward/backward rotation: 1/-1, 0 for none */
int rotate;                       /* direction of animation rotation */
int anim_params;                  /* bit mask for animation of single parameters */
int animate = 0;                  /* do automatic rotation */


#define PTILE_START 9
float compress_descr[] = {
  LF_VQ_TRAIN_SIZE, .04,
  LF_VQ_TILESIZE, 2, 2, 2, 2,
  LF_VQ_CODESIZE, 16384, 
  LF_VQ_PTILESIZE, 0, 1,
  LF_VQ_PTILESIZE, 1, 1,
  LF_VQ_PTILESIZE, 2, 1,
  LF_VQ_PTILESIZE, 3, 1,
  LF_VQ_PTILESIZE, 4, 1,
  LF_VQ_PTILESIZE, 5, 1,
  LF_VQ_PTILESIZE, 6, 1,
  LF_VQ_PTILESIZE, 7, 1,
  LF_VQ_PTILESIZE, 8, 1,
  LF_VQ_PTILESIZE, 9, 1,
  LF_NULL
};
bool_t vqP;


/*
** for drawing and handling the parameter display
*/
#define LINE_HEIGHT 13
#define CHAR_WIDTH   8
int win_height, win_width;
int params_height, params_left, params_y;
int draw_x, draw_y, draw_width, draw_height;
int dragging_param;
float dragging_val;
int params_descr_max;
static void init_params();
static void draw_params();

/*
 * set interpolation for selected or all variables
 */
static void setInterp(float *draw_descr, bool_t all)
{
  int i;

  for(i = 0; i < MAX_INTERPOLATE; i++) {
    draw_descr[i * 2 + 1] = (all || interp_params[i]) ? i : -1;
  }
}


/*
 * print and reset performance statistik
 */
int frame = 0;
time_t start;
float FramesPerSecond()
{
  int newt = time(NULL);

  return frame / (float)(newt - start);
}
void doTiming(void)
{
  int newt = time(NULL);

#ifdef STATS
  float clamp_ratio, user_timing, system_timing;
  float draw_reset[] = { LF_DRAW_STATS_RESET, LF_NULL };

  lfGetParameters(LF_DRAW, LF_DRAW_SCANLINE_CLAMP_RATIO, &clamp_ratio);
  lfGetParameters(LF_DRAW, LF_DRAW_TIMING_USER, &user_timing);
  lfGetParameters(LF_DRAW, LF_DRAW_TIMING_SYSTEM, &system_timing);
  printf("%d frames, clamp ratio %.3f, time user/system/total "
	 "%.3f/%.3f/%.3f => %.3f frames/s\n",
	 frame, clamp_ratio, user_timing, system_timing,
	 (float)(newt - start), frame / (float)(newt - start));
  lfSetParameters(LF_DRAW, draw_reset);
#else
  printf("%d frames, total time %.3f => %.3f frames/s\n",
	 frame, (float)(newt - start), frame / (float)(newt - start));
#endif
  start = newt;
  frame = 0;
}


/* Local helper functions */
void loadview();
void resetview();
void ParseArgs(int argc, char *argv[]);
void PrintUsage(char *programName);
LF_FileType DetermineFileType(char *fileName);
int CheckPow2(float *array, int first, int last);

/*
 * callback functions are used instead of event loop in main,
 * so variables previously local to main have been turned into globals
 */

void viewSetMenu(void);
void viewDisplayFunc(void);
void viewReshapeFunc(int width, int height);
void viewKeyboardFunc(unsigned char key, int x, int y);
void viewMotionFunc(int x, int y);
void viewMouseFunc(int button, int state, int x, int y);
void viewSpecialFunc(int key, int x, int y);
void viewMenuSelected(int value);

  /*
   * start main function now
   */
int
main(int argc, char *argv[])
{
  float read_descr[] = {
    LF_IGNORE,
    LF_IGNORE,
    LF_IGNORE,
    LF_READ_QUALITY, -1,
    LF_IGNORE,
    LF_READ_FORMAT, LF_RGBA,  /* hint: use RGBA as default because we will do rendering */
    LF_NULL,
  };
  LF_FileType fileType = LF_UNKNOWN;

  /* Put camera in default position */
  resetview();

  /* Set window size (if environment variable specifies it) */
  if  (getenv("LF_WIN_SIZE")) {
    sscanf(getenv("LF_WIN_SIZE"), "%ld%ld", &xsize, &ysize);
    glutInitWindowSize(xsize, ysize);
  }

  glutInit(&argc, argv);

  /* Parse the input arguments (verbose flag, input, etc) */
  ParseArgs(argc, argv);  

  if(noLight) {
    read_descr[0] = LF_READ_NO_LIGHT;
  }
  if(createImages) {
    read_descr[1] = LF_READ_CREATE_IMAGES;
  }
  if(noImages) {
    read_descr[2] = LF_READ_NO_IMAGES;
  }
  read_descr[4] = quality;
  if(uvFilter) {
    read_descr[5] = LF_READ_UVFILTER;
  }

  lfInitContext();

  /* Determine the type of file:
   * .lif, .lid, .gz, or other...
   */
  fileType = DetermineFileType(baseName);
  
  /* gunzip, if necessary */
  if (fileType == LF_GZ) {
    char cmdLine[1024];
    char *tempDir=NULL;
    /* Figure out if TEMP environment variable is set */
    tempDir = getenv("TEMP");
    if (tempDir != NULL) {
      sprintf(tempLif, "%s/current.lif", tempDir);
    } else {
      sprintf(tempLif, "current.lif");
    }

    /* gunzip, run DetermineFileType again... */
    sprintf(cmdLine, "gunzip -c %s > %s\n", baseName, tempLif);
    lfOutput("Executing: %s\n", cmdLine);
    if (system(cmdLine) != 0) {
      /* if gunzip fails, try gzip -d */
      sprintf(cmdLine, "gzip -dc %s > %s\n", baseName, tempLif);
      lfOutput("Attempting: %s\n", cmdLine);
      if (system(cmdLine) != 0) {
	/* If both fail, let user know */
	lfError("Unable to run gunzip/%s\n", cmdLine);
	exit(-1);
      }
    }
    lfOutput("Done!\n");
    baseName = strdup(tempLif);
    fileType = DetermineFileType(baseName);
  }


  if (fileType == LF_LIF) {
    /* Read in using new file format */
    lfOutput("Reading light field file %s\n", baseName);
    lfBegin(LF_FIELD, NULL);
    lfRead(LF_LIGHTFIELD, read_descr, baseName);
    lfEnd();
  } else if (fileType == LF_LID) {
    /* Read in using old file format */
    lfOutput("Reading/generating light field from slice images.\n");
    /* Read in the file, VQ compress it if requested */
    if(vqP) {
      read_descr[6] = LF_RGB;
    }
    lfBegin(LF_FIELD, NULL);
    lfRead(LF_SLICE_ST, read_descr, baseName);
    if(vqP) {
      lfCompress(LF_VQ, compress_descr);
    }
    lfEnd();
  } else {
    exit(-1);
  }

  /* If we gunzipped to create a temporary file, delete it. */
  if (!strcmp(baseName, tempLif)) {
    char cmdLine[1024];
    sprintf(cmdLine, "rm -f %s\n", tempLif);
    system(cmdLine);
    lfOutput("Removing %s \n", tempLif);
  }

  /* Now set up to start drawing */
  init_params();
  glutInitDisplayMode(GLUT_DOUBLE);
  glutCreateWindow("lifview");
  glutDisplayFunc(viewDisplayFunc);
  glutReshapeFunc(viewReshapeFunc);
  glutKeyboardFunc(viewKeyboardFunc);
  glutSpecialFunc(viewSpecialFunc);
  glutMouseFunc(viewMouseFunc);
  glutMotionFunc(viewMotionFunc);
  viewSetMenu();

  start = time(NULL);
  glutMainLoop();

  doTiming();

  return 0;
}


/* setup menu according to current settings */
void viewSetMenu(void)
{
  int menu, i;
  char buffer[120];
  LFParams *params = __LF_GET_PARAMS();
  LFRenderer *renderer = __LF_GET_RENDERER();
  LFView *view = __LF_GET_VIEW();
  bool_t no_orig = renderer->type == LF_RENDERER_NONE ||
    (renderer->flags & LFRF_NO_RENDERING);
  bool_t original = (view->force_rendering && !no_orig) ||
    noLight;

  menu = glutGetMenu();
  if(menu) {
    glutDestroyMenu(menu);
  }

  glutCreateMenu(viewMenuSelected);
  if(original) {
    if(!noLight) {
      glutAddMenuEntry("+ Original Rendering", MENU_RENDERING);
    } else {
      glutAddMenuEntry("Only Original Rendering", MENU_DUMMY);
    }
  } else {
    if(!no_orig) {
      glutAddMenuEntry("  Original Rendering", MENU_RENDERING);
    } else {
      glutAddMenuEntry("Only Light Field Rendering", MENU_DUMMY);
    }

    glutAddMenuEntry("", MENU_DUMMY);
    glutAddMenuEntry(!interp           ? "+ Point" : "  Point", MENU_POINT);
    glutAddMenuEntry((interp & UVLERP) ? "+ UV Interpolate" : "  UV Interpolate", MENU_UVINTER);
    glutAddMenuEntry((interp & STLERP) ? "+ ST Interpolate" : "  ST Interpolate", MENU_STINTER);
    for(i = 0; i < MAX_INTERPOLATE && i < params->num; i++) {
      sprintf(buffer, "%s Interpolate %s", interp_params[i] ? "+" : " ", params->params[i].descr);
      glutAddMenuEntry(buffer, MENU_INTERP(i));
    }
    glutAddMenuEntry("", MENU_DUMMY);
    glutAddMenuEntry(pixelzoom == 1    ? "+ Pixel Zoom 1"   : "  Pixel Zoom 1", MENU_ZOOM1);
    glutAddMenuEntry(pixelzoom == 2    ? "+ Pixel Zoom 2"   : "  Pixel Zoom 2", MENU_ZOOM2);
    glutAddMenuEntry(pixelzoom == 3    ? "+ Pixel Zoom 3"   : "  Pixel Zoom 3", MENU_ZOOM3);
  }

  if(renderer->sup_qualities) {
    glutAddMenuEntry("", MENU_DUMMY);
    for(i = 0; i < NUM_QUALITIES; i++) {
      if(renderer->sup_qualities & (1<<i)) {
	sprintf(buffer, "%s%s",
		quality == i ? "+ " : "  ",
		quality_names[i]);
	glutAddMenuEntry(buffer, MENU_QUALITY(i));
      }
    }
  }

  if(!original || renderer->sup_qualities) {
    glutAddMenuEntry("", MENU_DUMMY);
  }
  glutAddMenuEntry(!autoRefine       ? "+ Never Auto-Refine" : "  Never Auto-Refine", MENU_REFINE1);
  if(!original) {
    glutAddMenuEntry((autoRefine&REFINE_INTERP) ? "+ Auto-Refine Interpolate" : "  Auto-Refine Interpolate", MENU_REFINE2);
    glutAddMenuEntry((autoRefine&REFINE_ZOOM)   ? "+ Auto-Refine Zoom" : "  Auto-Refine Zoom", MENU_REFINE3);
  }
  if(renderer->sup_qualities) {
    glutAddMenuEntry((autoRefine&REFINE_QUAL) ? "+ Auto-Refine Quality" : "  Auto-Refine Quality", MENU_REFINEQ);
  }

  glutAddMenuEntry("", MENU_DUMMY);
  glutAddMenuEntry(lfIntensity ? "+ Intensity" : "   Intensity", MENU_INTENSITY);

  glutAddMenuEntry("", MENU_DUMMY);
  glutAddMenuEntry("Reset View", MENU_RESET);

  glutAddMenuEntry("", MENU_DUMMY);
  glutAddMenuEntry("Exit", MENU_EXIT);

  glutAttachMenu(GLUT_RIGHT_BUTTON);
}

/* refresh display */
void viewDisplayFunc(void)
{
  LFRenderer *renderer = __LF_GET_RENDERER();
  LFParams *params = __LF_GET_PARAMS();
  LFView *view = __LF_GET_VIEW();
  bool_t swap = TRUE;
  int i;

  glClear(GL_COLOR_BUFFER_BIT);

  loadview();
  if (!animate && refineredraw) {
    /* really something to improve? */
    if(refineredraw &
       ((view->force_rendering && renderer->type != LF_RENDERER_NONE) || noLight ? REFINE_QUAL : ~0)) {
      int tempzoom = pixelzoom;

      renderer->quality = (refineredraw & REFINE_QUAL) ? 0xFFFF : quality;
      /* mouse not down - high res. redraw */
      if (refineredraw & REFINE_ZOOM) pixelzoom = 1;	
      lfBegin(LF_FIELD, NULL);
      if (refineredraw & REFINE_INTERP) {
	setInterp(draw_descr[UVLERP + STLERP], TRUE);
	lfDraw(LF_COLOR, &(draw_descr[UVLERP + STLERP][0]), NULL);
      } else {
	setInterp(draw_descr[interp], FALSE);
	lfDraw(LF_COLOR, &(draw_descr[interp][0]), NULL);
      }
      lfEnd();
      pixelzoom = tempzoom;
    } else {
      swap = FALSE;
    }
    refineredraw = 0;
  } else {
    /* normal redraw */
    renderer->quality = quality;
    lfBegin(LF_FIELD, NULL);
    setInterp(draw_descr[interp], FALSE);
    lfDraw(LF_COLOR, &(draw_descr[interp][0]), NULL);
    lfEnd();
    /* If no motion, schedule higher-res redraw */
    if (autoRefine && !animate && !leftdown && !middledown) {
      refineredraw=autoRefine;
      glutPostRedisplay();
    } else {
      refineredraw=0;
    }
  }

  if(swap) {
    /* redraw parameters */
    draw_params();

    glutSwapBuffers();
    frame++;
  }

  if (animate || single_frame) {
    if  (animate > 0 || single_frame > 0)  {
      if  (rotate & ROTATEX)
	xrot += ROT_ANGLE_STEPS;
      if  (rotate & ROTATEY)
	yrot += ROT_ANGLE_STEPS;
      if  (rotate & ROTATEZ)
	zrot += ROT_ANGLE_STEPS;
      if  (xrot > MAX_ROT_ANGLE)
	animate = -1;
    }
    else {
      if  (rotate & ROTATEX)
	xrot -= ROT_ANGLE_STEPS;
      if  (rotate & ROTATEY)
	yrot -= ROT_ANGLE_STEPS;
      if  (rotate & ROTATEZ)
	zrot -= ROT_ANGLE_STEPS;
      if  (xrot < -MAX_ROT_ANGLE)
	animate = 1;
    }
    for(i = 0; i < params->num; i++) {
      if(anim_params & (1<<i)) {
	if(params->params[i].current == params->params[i].max) {
	  params->params[i].current = params->params[i].min;
	} else {
	  params->params[i].current++;
	}
      }
    }
    single_frame = 0;
    refineredraw = autoRefine;
    glutPostRedisplay();
  }
}

/* just keep track of state */
void viewMouseFunc(int button, int state, int x, int y)
{
  switch(button) {
  case GLUT_LEFT_BUTTON:
    leftdown = (state == GLUT_DOWN);
    break;
  case GLUT_MIDDLE_BUTTON:
    middledown = (state == GLUT_DOWN);
    break;
  }

  /* remember position for moving */
  lx = ((float) x) / (xsize*pixelzoom);
  ly = ((float) y) / (ysize*pixelzoom);

  if(y < params_height) {
    LFParams *params = &__LF_GET_CONTEXT()->field->params;
    /* GLUT counts from the top, so flip */
    y = win_height - y;
    dragging_param = (y - params_y) / LINE_HEIGHT;
    if(dragging_param < 0 || dragging_param >= params->num) {
      dragging_param = -1;
    } else {
      dragging_val   = params->params[dragging_param].current;
    }
  } else {
    dragging_param = -1;
  }

  /* Turn on auto-refine if mouse released */
  if (autoRefine && !leftdown && !middledown) {
    refineredraw = autoRefine;
    glutPostRedisplay();
  }
  else refineredraw = 0;
}

/* adapt GL viewport to new size and remember it */
void viewReshapeFunc(int width, int height)
{
  draw_x =
    draw_y = 0;
  draw_width = width;
  draw_height = height - params_height;
  xsize = (float)draw_width / pixelzoom;
  ysize = (float)draw_height / pixelzoom;
  win_height = height;
  win_width  = width;
  glutPostRedisplay();
}

/* check for user input */
static void check_animate()
{
  if(rotate || anim_params) {
    if(!animate)
      animate = 1;
  } else {
    animate = 0;
  }
  doTiming();
  glutPostRedisplay();
}
    
void viewKeyboardFunc(unsigned char key, int x, int y)
{
  switch(key) {
  case ' ':
    resetview();
    glutPostRedisplay();
    break;
  case '+':
    fov *= 1.1;
    glutPostRedisplay();
    break;
  case '-':
    fov /= 1.1;
    glutPostRedisplay();
    break;
  case '[':
    single_frame = -1;
    glutPostRedisplay();
    break;
  case ']':
    single_frame = 1;
    glutPostRedisplay();
    break;
  case '\e':
    doTiming();
    exit(0);
    break;
  case 'a':
    if(rotate) {
      rotate = 0;
    } else {
      rotate = ROTATEX | ROTATEY;
    }
    check_animate();
    break;
  case '\r':
  case '\n':
    doTiming();
    break;
  default:
    if('0' <= key && key <= '9') {
      anim_params ^= 1<<(key - '0');
      check_animate();
    }
  }
}
void viewSpecialFunc(int key, int x, int y)
{
  switch(key) {
  case GLUT_KEY_LEFT:
    params_left -= 8;
    if(params_left < 0) {
      params_left = 0;
    }
    viewReshapeFunc(win_width, win_height);
    break;
  case GLUT_KEY_RIGHT:
    params_left += 8;
    viewReshapeFunc(win_width, win_height);
    break;
  case GLUT_KEY_UP:
    params_height -= 8;
    if(params_height < 0) {
      params_height = 0;
    }
    viewReshapeFunc(win_width, win_height);
    break;
  case GLUT_KEY_DOWN:
    params_height += 8;
    viewReshapeFunc(win_width, win_height);
    break;
  }
  shift = (glutGetModifiers() & GLUT_ACTIVE_SHIFT);
}

/* change view */
void viewMotionFunc(int x, int y)
{  
  float newxobs, newyobs, newzobs;
  float mx, my, dx, dy;

  mx = ((float) x)/(xsize*pixelzoom);
  my = ((float) y)/(ysize*pixelzoom);
  dx = 2*(mx - lx)*xsize/DEFAULTSIZEX;
  dy = 2*(ly - my)*ysize/DEFAULTSIZEY;
  lx = mx;
  ly = my;

  /* changing a parameter? */
  if(dragging_param != -1) {
    if(leftdown || middledown) {
      LFParams *params = &__LF_GET_CONTEXT()->field->params;
      float delta = dy * (params->params[dragging_param].max - params->params[dragging_param].min);

      if(middledown) {
	delta /= 2;
      }
      dragging_val += delta;
      params->params[dragging_param].current = dragging_val;
      if(params->params[dragging_param].current > params->params[dragging_param].max) {
	dragging_val =
	  params->params[dragging_param].current = params->params[dragging_param].max;
      } else if(params->params[dragging_param].current < params->params[dragging_param].min) {
	dragging_val = 
	  params->params[dragging_param].current = params->params[dragging_param].min;
      }
      glutPostRedisplay();
    }
    return;
  }

  /* Mouse interactions -- pasted from Marc's viewer */
  if (leftdown && middledown) {
    /* Mouse Y maps to zooming toward/from center
     * mouse X does nothing
     * Movement slows down as you approach the far plane
     * Don't let observer reach far plane unless multifield
     */
    if (dy < 0) {
      xobs = xobs+(xobs-xcenter)*dy;
      yobs = yobs+(yobs-ycenter)*dy;
      zobs = zobs+(zobs-zcenter)*dy;
    }
    else {
      xobs = xcenter+(xobs-xcenter)/(1-dy);
      yobs = ycenter+(yobs-ycenter)/(1-dy);
      zobs = zcenter+(zobs-zcenter)/(1-dy);
    }
    /* Restore dy to original sign */
    dy = -dy;
  }
  else if (leftdown) {
    /* Mouse XY maps to longitude and lattitude
     * as observer orbits center at fixed distance
     * If shifted, suppress lattitude changes,
     * easing flyarounds of full-surround fields
     */
    float aold,anew;
    float rnew, rold;

    rold = sqrt((xobs-xcenter)*(xobs-xcenter)+
		(zobs-zcenter)*(zobs-zcenter));
    aold = fatan((xobs-xcenter)/(zobs-zcenter)) / M_PI*180.0;
    if (ABS(zobs-zcenter) < EPSILON) aold = 90.0 * 
				       SIGN(xobs-xcenter);
    else if (zobs<zcenter) aold = 180.0 + aold;
    anew = aold - dx * 90.0;
    
    xobs = xcenter + rold * fsin(anew/180.0*M_PI);
    zobs = zcenter + rold * fcos(anew/180.0*M_PI);

    rold = sqrt((xobs-xcenter)*(xobs-xcenter)+
		(yobs-ycenter)*(yobs-ycenter)+
		(zobs-zcenter)*(zobs-zcenter));
    newyobs = yobs-dy*rold;
    if (!shift) {
      yobs = newyobs;
    }
    rnew = sqrt((xobs-xcenter)*(xobs-xcenter)+
		(yobs-ycenter)*(yobs-ycenter)+
		(zobs-zcenter)*(zobs-zcenter));
    newxobs = (xobs-xcenter)*(rold/rnew)+xcenter;
    newyobs = (yobs-ycenter)*(rold/rnew)+ycenter;
    newzobs = (zobs-zcenter)*(rold/rnew)+zcenter;
    xobs = newxobs;
    if (!shift) {
      yobs = newyobs;
    }
    zobs = newzobs;
  }
  else if (middledown) {
    /* Mouse XY maps to translation of viewed image      */
    /* achieved by matched moves of observer and center  */
    /* perpendicular to the ray connecting them          */
    /* Such translation shows off horizontal parallax   */
    float newxcenter, newycenter, newzcenter;

    if (ABS(zcenter-zobs) > ABS(xcenter-xobs)) {
      /* Front and back views */
      newxcenter = xcenter+(zcenter-zobs)*dx/2;
      newzcenter = zcenter-(xcenter-xobs)*dx/2;
      newycenter = ycenter-ABS(zcenter-zobs)*dy/2;
      newzcenter = newzcenter-(ycenter-yobs)*dy/2;
      newxobs = xobs+(zcenter-zobs)*dx/2;
      newzobs = zobs-(xcenter-xobs)*dx/2;
      newyobs = yobs-ABS(zcenter-zobs)*dy/2;
      newzobs = newzobs-(ycenter-yobs)*dy/2;
    }
    else {
      /* Side views */
      newzcenter = zcenter-(xcenter-xobs)*dx/2;
      newxcenter = xcenter+(zcenter-zobs)*dx/2;
      newycenter = ycenter-ABS(xcenter-xobs)*dy/2;
      newxcenter = newxcenter-(ycenter-yobs)*dy/2;
      newzobs = zobs-(xcenter-xobs)*dx/2;
      newxobs = xobs+(zcenter-zobs)*dx/2;
      newyobs = yobs-ABS(xcenter-xobs)*dy/2;
      newxobs = newxobs-(ycenter-yobs)*dy/2;
    }
    xcenter = newxcenter;
    ycenter = newycenter;
    zcenter = newzcenter;
    /* If shifted, movement of observer is suppressed, */
    /* causing mouse XY to map to viewing direction at */
    /* the (fixed) current position, which exhibits    */
    /* no horizontal parallax but is often useful      */
    if (!shift) {
      xobs = newxobs;
      yobs = newyobs;
      zobs = newzobs;
    } 
  }
  glutPostRedisplay();
}

void viewMenuSelected(int value)
{
  int i;
  LFView *view = __LF_GET_VIEW();

  switch (value) {
  case MENU_RENDERING:
    view->force_rendering = !view->force_rendering;
    break;
  case MENU_POINT:
    interp = 0;
    break;
  case MENU_UVINTER:
    interp ^= UVLERP;
    break;
  case MENU_STINTER:
    interp ^= STLERP;
    break;
  case MENU_ZOOM1:
    pixelzoom = 1;
    break;
  case MENU_ZOOM2:
    pixelzoom = 2;
    break;
  case MENU_ZOOM3:
    pixelzoom = 3;
    break;
  case MENU_REFINE1:
    autoRefine = 0;
    break;
  case MENU_REFINE2:
    autoRefine ^= REFINE_INTERP;
    break;
  case MENU_REFINE3:
    autoRefine ^= REFINE_ZOOM;
    break;
  case MENU_REFINEQ:
    autoRefine ^= REFINE_QUAL;
    break;
  case MENU_INTENSITY:
    if(lfIntensity) {
      lfIntensity = NULL;
    } else {
      lfIntensity = &intensity;
    }
    viewReshapeFunc(win_width, win_height);
    break;
  case MENU_RESET:
    resetview();
    break;
  case MENU_EXIT:
    doTiming();
    exit(0);
    break;
  case MENU_DUMMY:
    return;
    break;
  default:
    for(i = 0; i < max(MAX_INTERPOLATE, NUM_QUALITIES); i++) {
      if(value == MENU_INTERP(i)) {
	interp_params[i] = interp_params[i] ? FALSE : TRUE;
	break;
      } else if(value == MENU_QUALITY(i)) {
	quality = i;
	break;
      }
    }
    break;
  }

  viewSetMenu();
  doTiming();
  glutPostRedisplay();
}



void setview(float fov, float aspect, float n, float f)
{
  float T[4][4];
#define RADIANS(a) ((3.14159/180)*(a))
  float cotfov2 = 1./tan(RADIANS(fov/2.));
  T[0][0] = cotfov2/aspect;
  T[0][1] = 0.;
  T[0][2] = 0.;
  T[0][3] = 0.;
    
  T[1][0] = 0.;
  T[1][1] = cotfov2;
  T[1][2] = 0.;
  T[1][3] = 0.;

  T[2][0] = 0.;
  T[2][1] = 0.;
  T[2][2] = -1.;
  T[2][3] = -1.;

  T[3][0] = 0.;
  T[3][1] = 0.;
  T[3][2] = -2*n;;
  T[3][3] = 0.;

  glMatrixMode(GL_PROJECTION);
  glLoadMatrixf((float *)T);
}

void loadview()
{
  LFView *view = &__LF_GET_CONTEXT()->view;

  /* in case renderer needs this instead of matrices */
  view->fov = fov;
  view->perspective = TRUE;
  view->nu = 8;
  view->n = .001;
  view->f = 1000.0;
  view->xobs = xobs;
  view->yobs = yobs;
  view->zobs = zobs;
  view->xcenter = xcenter;
  view->ycenter = ycenter;
  view->zcenter = zcenter;
  view->yrot = yrot;
  view->xrot = xrot;
  view->zrot = zrot;

  glViewport(draw_x,draw_y, draw_width, draw_height);
  setview(fov, xsize/(float)ysize, .001, 1000.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(xobs,yobs,zobs, xcenter,ycenter,zcenter, 0.0,1.0,0.0);
  glRotatef(yrot, 0.0,1.0,0.0);
  glRotatef(xrot, 1.0,0.0,0.0);
  glRotatef(zrot, 0.0,0.0,1.0);

  /* printf("rot %g y\n",yrot); */
}

void resetview()
{
  /* Move the viewer back to the starting point */
  xobs=RESETXOBS;
  yobs=RESETYOBS;
  zobs=RESETZOBS;
  xcenter = INITXCENTER;
  ycenter = INITYCENTER;
  zcenter = INITZCENTER;
  xrot = 0;
  yrot = 0;
  zrot = 0;
}

/*
** parameter handling
*/
static void init_params()
{
  int i;
  LFParams *params = &__LF_GET_CONTEXT()->field->params;
  LFRenderer *renderer = &__LF_GET_CONTEXT()->field->renderer;

  for(i = 0, params_descr_max = 0; i < params->num; i++) {
    int len = strlen(params->params[i].descr);

    params->params[i].current = params->params[i].def;
    if(len > params_descr_max)
      params_descr_max = len;
  }

  params_height = LINE_HEIGHT * (params->num + 1);
  params_left   = params_height * renderer->aspect;
}

static void print_at(int x, int y, char *text)
{
  glRasterPos2f(x,y);
  for(; *text; text++) {
    glutBitmapCharacter(GLUT_BITMAP_8_BY_13, *text);
  }
}
static void draw_params()
{
  int i;
  LFParams *params = &__LF_GET_CONTEXT()->field->params;
  LFRenderer *renderer = &__LF_GET_CONTEXT()->field->renderer;
  char buffer[200];

  glPushAttrib(GL_CURRENT_BIT);

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glTranslatef(-1.0, -1.0, 0.0);
  glScalef(2.0/win_width, 2.0/win_height, 1.0);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glViewport(0,0, win_width, win_height);
  glColor3f(1.0, 1.0, 1.0);

  params_y = win_height - params->num * LINE_HEIGHT;
  for(i = 0; i < params->num; i++) { 
    sprintf(buffer, "%d [%d/%d] - ", params->params[i].current, params->params[i].min, params->params[i].max);
    strncat(buffer, params->params[i].descr, sizeof(buffer));
    print_at(params_left, params_y + i * LINE_HEIGHT, buffer);
  }
  if(lfIntensity) {
    sprintf(buffer, "Intensity: %.2f", intensity);
  } else {
    sprintf(buffer, "frames/s: %.2f @ %dx%d (zoom %d)",
	    FramesPerSecond(),
	    (int)(win_width/pixelzoom),
	    (int)((win_height - params_height)/pixelzoom),
	    (int)pixelzoom);
  }
  print_at(params_left, params_y - LINE_HEIGHT, buffer);

  /* restrict window to upper left edge, coordinates from 0.0 to 1.0
     and draw parameters */
  glMatrixMode(GL_MODELVIEW);
  glTranslatef(0.0, win_height - params_height, 0.0);
  glScalef(params_left - 1, params_height - 1, 1.0);
  lfRDRenderParams(renderer);

  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  glPopAttrib();
}

void ParseArgs(int argc, char *argv[])
{
  char *progName = argv[0];
  int i;

  for(i = 0; i < sizeof(interp_params)/sizeof(interp_params[0]); i++) {
    interp_params[i] = TRUE;
  }
  
  argc--;
  argv++;

  while (argc > 0) {
    /* Verbose flag */
    if (!strcmp(argv[0], "-v")) {
      lfVerbose = 1;
      argc--;
      argv++;
    }
 
    /* Vector quantization compression */
    else if (!strcmp(argv[0], "-vq")) {
      vqP = 1;
      argv++;
      argc--;
    }
    /* -vqtrain:  size of the training set (fraction of entire set) */
    else if (!strcmp(argv[0], "-vqtrain")) {
      compress_descr[1] = atof(argv[1]);
      argv+=2;
      argc-=2;
      if (compress_descr[1] <=0) {
	fprintf(stderr, "Warning: vq training set must be greater than 0.\n");
	compress_descr[1] = .02;
      }
      if (compress_descr[1] > 1) {
	fprintf(stderr, 
		"Warning: vq training set cannot be greater than 1.\n");
	compress_descr[1] = 1.0;
      }	
      
    } 
    /* -vqtile: size (dimensions) of the vq tile */
    else if (!strcmp(argv[0], "-vqtile")) {
      if (argc < 5) PrintUsage(progName);
      compress_descr[3] = atoi(argv[1]);
      compress_descr[4] = atoi(argv[2]);
      compress_descr[5] = atoi(argv[3]);
      compress_descr[6] = atoi(argv[4]);
      argv += 5;
      argc -= 5;
      if (CheckPow2(compress_descr, 3, 6)) {
	fprintf(stderr, "Warning:  vq tile size must be a power of 2 "
		"in each dimension.\n");
      }
    }
    /* -vqtilep: size of tile in parameter #p */
    else if (!strcmp(argv[0], "-vqtilep")) {
      int pnum;
      float size;
      if (argc < 3) PrintUsage(progName);
      pnum = atoi(argv[1]);
      if(pnum < 0 || pnum >= LF_MAXPARAMS) {
	fprintf(stderr, "Error: -vqtilep parameter #%d not within [0, %d]\n", pnum, LF_MAXPARAMS - 1);
	exit(1);
      }
      size = atoi(argv[2]);
      if(CheckPow2(&size, 0, 0)) {
	fprintf(stderr, "Warning:  vq tile size must be a power of 2 "
		"for each parameter.\n");
      }
      compress_descr[PTILE_START + pnum * 3 + 2] = size;
      argv += 3;
      argc -= 3;
    }
    /* -vqcode: size of the codebook (number of entries */
    else if (!strcmp(argv[0], "-vqcode")) {
      compress_descr[8] = atoi(argv[1]);
      argv+= 2;
      argc-= 2;
      /* For now, disallow codebooks smaller than 257.
       * The code doesn't handle 1-byte code indices correctly yet.
       */
      if (compress_descr[8] < 257) {
	fprintf(stderr, "Warning: vqcodebook too small!\n");
	fprintf(stderr, "Warning: Current minimum codebook size: 257\n");
	compress_descr[8] = 257;
      }
      if (compress_descr[8] > 65536) {
	fprintf(stderr, "Warning: vqcodebook is very large. "
		"This may take a while...\n");
      }
    }

    /* zoom */
    else if (!strcmp(argv[0], "-zoom")) {
      if (argc < 2) PrintUsage(progName);
      pixelzoom = atoi(argv[1]);
      /* error checking */
      if (pixelzoom < 1 || pixelzoom > 100) {
	lfError("Invalid zoom scale... using default\n");
	pixelzoom = 2.0;
      }
      argc -= 2;
      argv += 2;
    } 
    
    /* Interpolation mode */
    else if (!strcmp(argv[0], "-lerp")) {
      if (argc < 2) PrintUsage(progName);
      if (!strcmp(argv[1], "none")) {
	interp = 0;
      } else if (!strcmp(argv[1], "uv")) {
	interp = UVLERP;
      } else if (!strcmp(argv[1], "st")) {
	interp = STLERP;
      } else if (!strcmp(argv[1], "uvst")) {
	interp = STLERP + UVLERP;
      } else {
	lfError("Invalid lerp mode: %s\n", argv[1]);
	PrintUsage(progName);
      }
      argc -= 2;
      argv += 2;
    }

    /* Auto-Refine */
    else if (!strcmp(argv[0], "-refine")) {
      if (argc < 2) PrintUsage(progName);
      if (!strcmp(argv[1], "all")) {
	autoRefine = REFINE_INTERP + REFINE_ZOOM + REFINE_QUAL;
      } else if (!strcmp(argv[1], "uvst")) {
	autoRefine = REFINE_INTERP;
      } else if (!strcmp(argv[1], "zoom")) {
	autoRefine = REFINE_ZOOM;
      } else if (!strcmp(argv[1], "off")) {
	autoRefine = 0;
      } else {
	lfError("Invalid auto-refine mode: %s\n", argv[1]);
	PrintUsage(progName);
      }
      argc -= 2;
      argv += 2;
    }

    /* source selection */
    else if (!strcmp(argv[0], "-create_images")) {
      createImages = 1;
      argc--;
      argv++;
    } else if (!strcmp(argv[0], "-uvfilter")) {
      uvFilter = 1;
      argc--;
      argv++;
    } else if (!strcmp(argv[0], "-no_light")) {
      noLight = 1;
      argc--;
      argv++;
    } else if (!strcmp(argv[0], "-no_images")) {
      noImages = 1;
      argc--;
      argv++;
    }
    
    /* default intensity */
    else if (!strcmp(argv[0], "-intensity")) {
      lfIntensity = &intensity;
      argc--;
      argv++;
    }

    /* default quality */
    else if (!strcmp(argv[0], "-quality")) {
      quality = atoi(argv[1]);
      argc -= 2;
      argv += 2;
    }

    /* filename */
    else if (strncmp(argv[0], "-", 1)) {
      if (baseName == NULL) {
	baseName = argv[0];
      } else {
	lfError("Ignoring extra file %s\n", baseName);
      } 
      argc--;
      argv++;
    } 
    
    /* default */
    else {
      PrintUsage(progName);
      exit(-1);
    }
  }

  /* Check to make sure baseName has been set */
  if (baseName == NULL) {
    PrintUsage(progName);
  }
}
  

void PrintUsage(char *programName)
{
  fprintf(stderr, 
	  "lifview for OpenGL v.1.0\n"
	  "Usage: %s [-v] [-zoom <z>] [-lerp <mode>] \n"
	  "               [-refine <mode>] [-create_images] [-no_light]\n"
	  "               <lightfield>\n"
	  "\n"
	  "     -v              Verbose on\n"
	  "\n"
	  "     -zoom <z>       Zoom factor for display window\n"
	  "                     (The actual window size will be x*z, y*z)\n"
	  "\n"
	  "     -lerp uvst      Start with full uvst interpolation filter\n"
	  "     -lerp uv        Start with only uv interpolation (default)\n"
	  "     -lerp st        Start with only st interpolation\n"
	  "     -lerp none      Start with point sampling\n"
	  "\n"
	  "     -refine all     Auto-refine image when viewer isn't moving\n"
	  "                     (this refines both uvst and zoom)\n"
	  "     -refine uvst    Auto-refine uvst interpolation only\n"
	  "     -refine zoom    Auto-refine pixel zoom only\n"
	  "     -refine off     Never auto-refine image\n"
	  "\n"
	  "     -intensity      Start with enabled calculation and display of intensity\n"
	  "\n"
	  "     -quality n      Start with this rendering quality:\n"
	  "                     1 - " QUALITY_CRUDE "\n"
	  "                     2 - " QUALITY_LOW "\n"
	  "                     3 - " QUALITY_MEDIUM "\n"
	  "                     4 - " QUALITY_HIGH "\n"
	  "\n"
	  "     -create_images  When reading a .lid and generating the light\n"
	  "                     field using the defined renderer, then also\n"
	  "                     create the image files.\n"
	  "     -no_light       Instead of initialising both the light field and\n"
	  "                     the (optional) renderer, rely on renderer alone.\n"
	  "\n"
	  "     -no_images      The renderer is used even if an image exists.\n"
	  "                     Images are still created if requested.\n"
	  "\n"
	  "     <lightfield>    This can be the name of either a .lif\n"
	  "                     file or a .lid file.  If it is a .lid\n"
	  "                     file, it will tell lifview where to\n"
	  "                     find the raw .rgb images.\n"
	  "\n"
	  , programName);
  
  exit(-1);
}

LF_FileType DetermineFileType(char *fileName)
{
  LF_FileType type = LF_UNKNOWN;
  char buffer[20];
  FILE *inFile = NULL;
  char *suffixPtr, *tPtr;

  /* Open the input file */
#ifdef sgi
  inFile = fopen(fileName, "rb");
#else /* WIN32 */
  inFile = fopen(fileName, "r+b");
#endif 
  if (!inFile) {
    lfError("Cannot open file %s\n", fileName);
    exit(-1);
  }
  
  /* Get the first 8 characters of the file to determine its type */
  fgets(buffer, 9, inFile);
  
  switch(buffer[0]) {
  case 'L':  /* LIF1.0 */
    if (!strncmp(buffer, "LIF", 3)) {
      if (!strncmp(buffer, "LIF1.0", 6)) {
	type = LF_LIF;
      } else {
	lfError("Cannot handle LIF version number \"%s\".\n"
		"This is lifview 1.0.\n", &(buffer[3]));
	type = LF_LIFBADVER;
      }
      
    } else {
      type = LF_UNKNOWN;
    }
    break;
  case '#': /* # LIF1.0 */
    type = LF_LID;
    break;

  case 31:  /* gzip magic byte */
    type = LF_GZ;
    break;
  default:
    type = LF_UNKNOWN;
    break;
  }

  /* If the type is still unknown, look at the suffix of the
   * filename, and try to use it if possible, or at least print
   * a reasonable error message. 
   */
  if (type == LF_UNKNOWN) {
    /* Figure out the suffix */
    suffixPtr = fileName;
    /* Make suffixPtr point to character after the last . */
    for (tPtr = fileName; tPtr < (fileName + strlen(fileName)); tPtr++) {
      if (*tPtr == '.') suffixPtr = tPtr+1;
    }
    
    /* Now test it against known types */
    if (!strcmp(".lif", suffixPtr) || !strcmp(".LIF", suffixPtr)) {
      lfError("Bad .lif file, does not start with \"LIF1.0\".\n");
    }
    else if (!strcmp(".gz", suffixPtr) || !strcmp(".GZ", suffixPtr)) {
      lfError("Bad .gz file, does not start with magic characters.\n");
    }
    else if (!strcmp(".lid", suffixPtr) || !strcmp(".LID", suffixPtr)) {
      /* This is ok, I guess.  They might've stripped out the comments. */
      type = LF_LID;
    } else {
      lfError("Unrecognized input format for file \"%s\".\n", fileName);
    }
  }

  /* Close up file, return */
  fclose(inFile);
  return(type);
}

/* This function makes sure that every element of the array between
 * first and last (inclusive) is a power of 2, between 1 and 256.
 */
int CheckPow2(float *array, int first, int last)
{
  int i;
  int pow;
  int errorP = 0;

  for (i=first; i <= last; i++) {
    /* find pow, the ceiling of log (base 2) of array[i]. */
    for (pow = 0; pow <= 8; pow++) {
      if (array[i] <= (1 << pow)) break;
    }	
    if (array[i] != (1 << pow)) {
      errorP = 1;
      array[i] = (1 << pow);
    }
  }
  return errorP;
}
